from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Union

import cappa
from tests.utils import Backend, backends, invoke


@dataclass
class Subcommand:
    ...

    def __call__(self, command: cappa.Command[Any]):
        return command.real_name()


@dataclass
class TopLevelCommand:
    subcmd: cappa.Subcommands[Union[Subcommand, None]] = None

    def __call__(self, command: cappa.Command[Any]):
        return command.real_name()


@backends
def test_command_as_dependency(backend: Backend):
    result = invoke(TopLevelCommand, backend=backend)
    assert result == "top-level-command"

    result = invoke(TopLevelCommand, "subcommand", backend=backend)
    assert result == "subcommand"



from __future__ import annotations

from dataclasses import dataclass
from typing import Any
from unittest.mock import patch

import pytest
from rich.console import Console
from rich.text import Text
from typing_extensions import Annotated

import cappa
from tests.utils import parse


def test_line_wraps_correctly_with_terminal_escape_codes(capsys: Any):
    @dataclass
    class Args:
        a: Annotated[str, cappa.Arg(short=True)]
        b: Annotated[str, cappa.Arg(short=True)]
        c: Annotated[str, cappa.Arg(short=True)]
        d: Annotated[str, cappa.Arg(short=True)]
        e: Annotated[str, cappa.Arg(short=True)]
        f: Annotated[str, cappa.Arg(short=True)]
        g: Annotated[str, cappa.Arg(short=True)]
        i: Annotated[str, cappa.Arg(short=True)]

    columns = 80
    env = {
        "FORCE_COLOR": "true",
        "COLUMNS": str(columns),
    }
    with patch("os.environ", new=env), pytest.raises(cappa.HelpExit) as e:
        output = cappa.Output(output_console=Console(force_terminal=True))
        parse(Args, "--help", output=output)

    assert e.value.code == 0

    out = capsys.readouterr().out
    plain, *_ = Text.from_ansi(out).plain.splitlines()

    expected = "Usage: args -a A -b B -c C -d D -e E -f F -g G -i I [-h] [--completion"
    assert len(expected) < columns
    assert plain == expected



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import pytest
from typing_extensions import Annotated

import cappa
from cappa.output import Exit
from tests.utils import (
    Backend,
    CapsysOutput,
    backends,
    parse,
    terminal_width,
)


@dataclass
class Args:
    verbose: Annotated[
        int,
        cappa.Arg(short="-v", help="All help is included."),
        cappa.Arg(long="--verbosity", help="But the default is not duplicated"),
    ] = 0


@backends
def test_help(backend: Backend, capsys: Any):
    with terminal_width(), pytest.raises(Exit):
        parse(Args, "--help", backend=backend)

    out = CapsysOutput.from_capsys(capsys)
    assert "All help is included. But the default is not duplicated" in out.stdout
    assert "(Default: 0)" in out.stdout



from __future__ import annotations

import dataclasses
import re
from collections import deque
from functools import cached_property
from typing import (
    Any,
    Callable,
    Generic,
    Hashable,
    Iterable,
    cast,
)

from cappa.arg import Arg, ArgAction, ArgActionType, FinalArg
from cappa.command import Alias, Command, FinalCommand
from cappa.completion.types import Completion, FileCompletion
from cappa.help import format_args, format_subcommand_names
from cappa.invoke.base import fulfill_deps
from cappa.output import Exit, HelpExit, Output
from cappa.subcommand import FinalSubcommand
from cappa.typing import T

negative_number = re.compile(r"^-\d+$|^-\d*\.\d+$")


class BadArgumentError(RuntimeError):
    def __init__(
        self,
        message: str,
        *,
        value: Any,
        command: FinalCommand[Any],
        arg: Arg[Any] | FinalSubcommand | None = None,
    ) -> None:
        super().__init__(message)
        self.value = value
        self.command = command
        self.arg = arg


@dataclasses.dataclass
class HelpAction(RuntimeError):
    command: FinalCommand[Any]
    command_name: str

    @classmethod
    def from_parse_state(cls, parse_state: ParseState, command: FinalCommand[Any]):
        raise cls(command, parse_state.prog)


@dataclasses.dataclass
class VersionAction(RuntimeError):
    version: Arg[Any]

    @classmethod
    def from_arg(cls, arg: Arg[Any]):
        raise cls(arg)


class CompletionAction(RuntimeError):
    def __init__(
        self,
        *completions: Completion | FileCompletion,
        value: str = "complete",
        arg: Arg[Any] | None = None,
    ) -> None:
        self.completions = completions
        self.value = value
        self.arg = arg

    @classmethod
    def from_value(cls, value: Value[str], arg: Arg[Any]):
        raise cls(value=value.value, arg=arg)


def backend(
    command: FinalCommand[T],
    argv: list[str],
    output: Output,
    prog: str,
    provide_completions: bool = False,
) -> tuple[Any, FinalCommand[T], dict[str, Any]]:
    context = ParseContext.from_command(command)
    parse_state = ParseState.from_command(
        argv, command, output=output, provide_completions=provide_completions
    )

    try:
        try:
            parse(parse_state, context)
        except HelpAction as e:
            raise HelpExit(
                e.command.help_formatter.long(e.command, e.command_name),
                code=0,
                prog=parse_state.prog,
            )
        except VersionAction as e:
            raise Exit(cast(str, e.version.value_name), code=0, prog=parse_state.prog)
        except BadArgumentError as e:
            if parse_state.provide_completions and e.arg:
                completions = e.arg.completion(e.value) if e.arg.completion else []
                raise CompletionAction(*completions)

            raise Exit(str(e), code=2, prog=parse_state.prog, command=e.command)
    except CompletionAction as e:
        from cappa.completion.base import execute, format_completions

        if provide_completions:
            completions = format_completions(*e.completions)
            raise Exit(completions, code=0)

        execute(command, prog, e.value, cast(Arg[Any], e.arg), output=output)

    if provide_completions:
        raise Exit(code=0)

    return (parse_state, parse_state.current_command or command, context.result)


@dataclasses.dataclass
class ParseState:
    """The overall state of the argument parse."""

    args: ArgCollection
    command_stack: list[FinalCommand[Any]]
    output: Output
    provide_completions: bool = False

    @classmethod
    def from_command(
        cls,
        argv: list[str],
        command: FinalCommand[Any],
        output: Output,
        provide_completions: bool = False,
    ):
        arg_stream = ArgCollection(deque(argv), provide_completions=provide_completions)
        return cls(
            arg_stream,
            command_stack=[command],
            output=output,
            provide_completions=provide_completions,
        )

    @property
    def current_command(self):
        return self.command_stack[-1]

    @property
    def prog(self):
        return " ".join(c.real_name() for c in self.command_stack)

    def push_command(self, command: FinalCommand[Any]):
        self.command_stack.append(command)


@dataclasses.dataclass
class ParseContext:
    """The parsing context specific to a command."""

    command: FinalCommand[Any]
    arguments: deque[FinalArg[Any] | FinalSubcommand]
    missing_options: set[str]
    arguments_by_field_name: dict[str, list[FinalArg[Any]]]
    arguments_by_value_name: dict[str, FinalArg[Any]]
    propagated_options: set[str]
    parent_context: ParseContext | None = None
    exclusive_args: dict[str, FinalArg[Any]] = dataclasses.field(
        default_factory=lambda: {}
    )

    result: dict[str, Any] = dataclasses.field(default_factory=lambda: {})

    @classmethod
    def from_command(
        cls,
        command: FinalCommand[Any],
        parent_context: ParseContext | None = None,
    ) -> ParseContext:
        arguments_by_field_name: dict[str, list[FinalArg[Any]]] = {}
        arguments_by_value_name: dict[str, FinalArg[Any]] = {}
        propagated_options: set[str] = set()
        unique_field_names: set[str] = set()

        def add_option_names(arg: FinalArg[Any]):
            for opts in (arg.short, arg.long):
                if not opts:
                    continue

                for key in opts:
                    if key in arguments_by_value_name:
                        raise ValueError(f"Conflicting option string: {key}")

                    arguments_by_value_name[key] = arg

        native_command_field_names: set[str] = set()
        for arg in command.options:
            field_name = arg.field_name

            if arg.action not in ArgAction.meta_actions():
                unique_field_names.add(field_name)

            native_command_field_names.add(field_name)
            arguments_by_field_name.setdefault(field_name, []).append(arg)
            add_option_names(arg)

        for arg in command.propagated_arguments:
            field_name = arg.field_name

            if field_name in native_command_field_names:
                continue

            propagated_options.add(field_name)
            arguments_by_field_name.setdefault(field_name, []).append(arg)
            add_option_names(arg)

        arguments = deque(command.positional_arguments)

        return cls(
            command=command,
            parent_context=parent_context,
            arguments_by_field_name=arguments_by_field_name,
            arguments_by_value_name=arguments_by_value_name,
            propagated_options=propagated_options,
            arguments=arguments,
            missing_options=unique_field_names,
        )

    @cached_property
    def propagated_context(self) -> dict[str, ParseContext]:
        parent_context = (
            self.parent_context.propagated_context if self.parent_context else {}
        )
        self_options = {o.field_name: o for o in self.command.options if o.propagate}
        self_context: dict[str, ParseContext] = dict.fromkeys(self_options, self)
        return {**parent_context, **self_context}

    def next_argument(self):
        return self.arguments.popleft()

    def resolve_context(self, field_name: str, option: RawOption | None = None):
        context = self
        if option and field_name in self.propagated_options:
            context = self.propagated_context[field_name]
        return context

    def set_result(
        self,
        field_name: str,
        value: Any,
        option: RawOption | None = None,
        has_value: bool = True,
    ):
        if option and field_name in self.missing_options:
            self.missing_options.remove(field_name)

        if has_value:
            self.result[field_name] = value

    def push(self, command: FinalCommand[Any], name: str) -> ParseContext:
        nested_context = ParseContext.from_command(command, parent_context=self)
        nested_context.result["__name__"] = name
        return nested_context


@dataclasses.dataclass
class ArgCollection:
    """Represent the pending parse state of remaining unparsed arguments.

    Whether a given fragment to be parsed is considered an Option or an Argument
    is context dependent (e.x. following `--`) and it simplifies code downstream
    if that decision is being made on-demand rather than eagerly in a way that
    needs to be corrected at the use-sites.
    """

    argv: deque[str]
    provide_completions: bool = False
    encountered_double_dash: bool = False
    pending_args: deque[RawOption | RawArg] = dataclasses.field(
        default_factory=lambda: deque()
    )

    def has_values(self) -> bool:
        return bool(self.argv) or bool(self.pending_args)

    def peek_value(self, context: ParseContext) -> RawArg | RawOption | None:
        next_value = self.next(context)
        if next_value is None:
            return None

        self.pending_args.appendleft(next_value)
        return next_value

    def next(self, context: ParseContext) -> RawArg | RawOption | None:
        if self.pending_args:
            return self.pending_args.popleft()

        if self.argv:
            next_value = self.argv.popleft()

            if self.encountered_double_dash:
                item: RawArg | RawOption | None = RawArg(next_value)
            else:
                item = RawArg.from_str(
                    next_value, provide_completions=self.provide_completions
                )

            if item is None:
                self.encountered_double_dash = True
                return None

            if isinstance(item, RawOption) and not item.is_long:
                item = self.generate_virtual_args(item, context.arguments_by_value_name)
            return item

        return None

    def generate_virtual_args(
        self, arg: RawOption, options: dict[str, Any]
    ) -> RawOption | RawArg:
        """Produce "virtual" options from short (potentially concatenated) options.

        Examples:
            -abc -> -a, -b, -c
            -c0 -> -c 0
            -abc0 -> -a, -b, -c, 0
        """
        result: list[RawOption | RawArg] = []

        partial_arg = ""
        remaining_arg = arg.name[1:]
        while remaining_arg:
            partial_arg += remaining_arg[0]
            remaining_arg = remaining_arg[1:]

            option_name = f"-{partial_arg}"

            option = options.get(option_name)
            if option:
                result.append(RawOption(option_name, value=arg.value))
                partial_arg = ""

                # An option which requires consuming further arguments should consume
                # the rest of the concatenated character sequence as its value.
                if option.num_args.n:
                    partial_arg = remaining_arg
                    break

        if not result:
            # i.e. -p, where -p is not a real short option. It will get skipped above.
            option = RawOption(arg.name, value=arg.value)
            return option.to_negative_number() or option

        if partial_arg:
            result.append(RawArg(partial_arg))

        option, *virtual_args = result
        self.pending_args.extend(virtual_args)
        return option


@dataclasses.dataclass
class RawArg:
    raw: str

    @classmethod
    def from_str(
        cls, arg: str, *, provide_completions: bool = False
    ) -> RawArg | RawOption | None:
        skip = arg == "--" and not provide_completions
        if skip:
            return None

        is_option = arg and arg[0] == "-" and (provide_completions or len(arg) > 1)

        if is_option:
            return RawOption.from_str(arg)

        return cls(arg)


@dataclasses.dataclass
class RawOption:
    name: str
    is_long: bool = True
    value: str | None = None

    @classmethod
    def from_str(cls, arg: str) -> RawOption:
        is_long = arg.startswith("--")
        is_explicit = "=" in arg

        name = arg
        value = None
        if is_explicit:
            name, value = arg.split("=", 1)
        return cls(name=name, is_long=is_long, value=value)

    def to_negative_number(self) -> RawArg | None:
        if re.match(negative_number, self.name):
            return RawArg(self.name)
        return None


def parse(parse_state: ParseState, context: ParseContext) -> None:
    while True:
        while isinstance(parse_state.args.peek_value(context), RawOption):
            arg = cast(RawOption, parse_state.args.next(context))
            parse_option(parse_state, context, arg)

        parse_args(parse_state, context)

        if not parse_state.args.has_values():
            break

    # Options are not explicitly iterated over because they can occur multiple times non-contiguouesly.
    # So instead we check afterward, if there are any missing which we haven't yet fulfilled.
    required_missing_options = [
        arg
        for opt_name in sorted(context.missing_options)
        for arg in context.arguments_by_field_name[opt_name]
        if arg.required
    ]
    if required_missing_options:
        names = ", ".join([opt.names_str("/") for opt in required_missing_options])
        raise BadArgumentError(
            f"The following arguments are required: {names}",
            value="",
            command=parse_state.current_command,
            arg=required_missing_options[0],
        )


def parse_option(
    parse_state: ParseState, context: ParseContext, raw: RawOption
) -> None:
    if raw.name not in context.arguments_by_value_name:
        possible_options: dict[str, FinalArg[Any]] = {
            name: arg
            for name, arg in context.arguments_by_value_name.items()
            if name.startswith(raw.name)
        }

        if parse_state.provide_completions:
            options: list[Completion] = []
            for name, option in possible_options.items():
                rendered_help = str(
                    format_args(
                        parse_state.output.output_console,
                        context.command.help_formatter,
                        option,
                    )
                ).strip()
                completion = Completion(name, help=rendered_help, arg=option)
                options.append(completion)

            raise CompletionAction(*options)

        message = f"Unrecognized arguments: {raw.name}"
        if possible_options:
            message += f" (Did you mean: {', '.join(possible_options.keys())})"

        raise BadArgumentError(
            message, value=raw.name, command=parse_state.current_command
        )

    arg = context.arguments_by_value_name[raw.name]

    consume_arg(parse_state, context, arg, raw)


def parse_args(parse_state: ParseState, context: ParseContext) -> None:
    while context.arguments:
        peeked = parse_state.args.peek_value(context)
        if isinstance(peeked, RawOption):
            break

        arg = context.next_argument()

        if isinstance(arg, FinalSubcommand):
            consume_subcommand(parse_state, context, arg)
        else:
            consume_arg(parse_state, context, arg)
    else:
        value = parse_state.args.peek_value(context)
        if value is None or isinstance(value, RawOption):
            return

        raw_values: list[str] = []
        while parse_state.args.peek_value(context):
            next_val = parse_state.args.next(context)
            if not isinstance(next_val, RawArg):
                break

            raw_values.append(next_val.raw)

        raise BadArgumentError(
            f"Unrecognized arguments: {', '.join(raw_values)}",
            value=raw_values,
            command=parse_state.current_command,
        )


def consume_subcommand(
    parse_state: ParseState, context: ParseContext, arg: FinalSubcommand
) -> Any:
    value = parse_state.args.next(context)
    if value is None:
        if not arg.required:
            return

        raise BadArgumentError(
            f"A command is required: {{{format_subcommand_names(arg.names())}}}",
            value="",
            command=parse_state.current_command,
            arg=arg,
        )

    assert isinstance(value, RawArg), value
    canonical = arg.resolve_name(value.raw)
    if canonical is None:
        message = f"Invalid command '{value.raw}'"
        possible_values = [
            name for name in arg.all_visible_names() if name.startswith(value.raw)
        ]
        if possible_values:
            message += f" (Did you mean: {format_subcommand_names(possible_values)})"

        raise BadArgumentError(
            message,
            value=value.raw,
            command=parse_state.current_command,
            arg=arg,
        )

    command = arg.options[canonical]
    check_deprecated(parse_state, command)

    if value.raw != canonical:
        _, alias = arg.alias_map[value.raw]
        _warn_deprecated_alias(parse_state, alias)

    parse_state.push_command(command)
    nested_context = context.push(command, canonical)

    parse(parse_state, nested_context)

    context.result[arg.field_name] = nested_context.result


def iter_arg_values(
    context: ParseContext,
    parse_state: ParseState,
    num_args: int,
    option: RawOption | None,
) -> Iterable[str]:
    """Collect argument values from the parse state."""
    # If option has explicit value (e.g., --opt=val), yield it and stop
    if option and option.value:
        yield option.value
        return

    remaining = num_args
    while remaining:
        peeked_value = parse_state.args.peek_value(context)
        if peeked_value is None:
            break

        if isinstance(peeked_value, RawOption):
            break

        next_value = peeked_value.raw

        parse_state.args.next(context)
        yield next_value

        # num_args == -1 means unbounded, so remaining will always be truthy
        remaining -= 1


def arg_bypasses_action(
    arg: FinalArg[Any],
    values: list[str],
    expected_count: int,
    option: RawOption | None,
    parse_state: ParseState,
) -> bool:
    """Check whether the current argument is in a state that bypasses action processing.

    This is mainly to handle falling back to the default value when an expected one is
    missing.
    """
    is_positional = not option

    is_option_collecting_values = option and not option.value

    is_single_value = expected_count == 1
    is_fixed_count = expected_count > 0
    is_unbounded = expected_count < 0

    has_value = bool(values)
    has_expected_value = is_fixed_count and len(values) == expected_count
    expects_value = option or arg.required

    if has_expected_value:
        return False

    if is_single_value:
        if not has_value and expects_value:
            raise BadArgumentError(
                f"Option '{arg.value_name}' requires an argument",
                value="",
                command=parse_state.current_command,
                arg=arg,
            )

        return not has_value

    is_unbounded_empty = is_unbounded and not has_value
    if not is_fixed_count and not is_unbounded_empty:
        return False

    is_optional_positional_without_value = (
        is_positional and not arg.required and not has_value
    )

    is_unbounded_optional_without_value = (
        is_unbounded_empty and not arg.required and is_option_collecting_values
    )

    if (
        not is_optional_positional_without_value
        and not is_unbounded_optional_without_value
    ):
        names_str = arg.names_str("/")
        count_desc = "at least one" if is_unbounded else expected_count
        message = (
            f"Argument '{names_str}' requires {count_desc} values, found {len(values)}"
        )
        if values:
            quoted = [f"'{v}'" for v in values]
            message += f" ({', '.join(quoted)} so far)"
        raise BadArgumentError(
            message,
            value=values,
            command=parse_state.current_command,
            arg=arg,
        )

    return is_optional_positional_without_value


def check_exclusive_group(
    arg: FinalArg[Any],
    context: ParseContext,
    result: Any,
    parse_state: ParseState,
) -> None:
    """Check if arg violates exclusive group constraints."""
    group = arg.group
    if not group.exclusive:
        return

    exclusive_arg = context.exclusive_args.get(group.id)

    if exclusive_arg and exclusive_arg != arg:
        raise BadArgumentError(
            f"Argument '{arg.names_str('/')}' is not allowed with argument"
            f" '{exclusive_arg.names_str('/')}'",
            value=result,
            command=parse_state.current_command,
            arg=arg,
        )

    context.exclusive_args[group.id] = arg


def _resolve_context(
    context: ParseContext,
    field_name: str,
    option: RawOption | None,
    arg: FinalArg[Any],
) -> tuple[ParseContext, bool]:
    resolved = context.resolve_context(field_name, option)
    if arg.destructure:
        nested = resolved.result.setdefault(arg.destructure.field_name, {})
        resolved = dataclasses.replace(resolved, result=nested)
        return resolved, True
    return resolved, arg.has_value


def consume_arg(
    parse_state: ParseState,
    context: ParseContext,
    arg: FinalArg[Any],
    option: RawOption | None = None,
) -> Any:
    field_name = arg.field_name

    # Determine how many values to collect
    normalized_num_args = arg.num_args
    expected_count = normalized_num_args.n
    if ArgAction.is_non_value_consuming(arg.action):
        expected_count = 0

    values = list(iter_arg_values(context, parse_state, expected_count, option))

    # Argument whose value is optional, without a value: falls back to NumArgs.default. Distinct from
    # an optional argument, who's value would be unrecorded and apply the default during mapping.
    if option is not None and not normalized_num_args.required and not values:
        check_exclusive_group(arg, context, normalized_num_args.default, parse_state)
        resolved_context, has_value = _resolve_context(context, field_name, option, arg)
        resolved_context.set_result(
            field_name, normalized_num_args.default, option, has_value
        )
        check_deprecated(parse_state, arg, option)
        return

    if arg_bypasses_action(arg, values, expected_count, option, parse_state):
        return

    # Convert single-value args to scalar
    result: str | list[str] = values[0] if expected_count == 1 else values

    # Handle completions for args with values
    if parse_state.provide_completions and not parse_state.args.has_values() and values:
        if arg.completion:
            completions: list[Completion] | list[FileCompletion] = arg.completion(
                result
            )
        else:
            completions = [FileCompletion(values[-1])]
        raise CompletionAction(*completions)

    check_exclusive_group(arg, context, result, parse_state)

    action_handler = determine_action_handler(arg.action)
    resolved_context, has_value = _resolve_context(context, field_name, option, arg)

    fulfilled_deps: dict[Hashable, Any] = {
        FinalCommand: parse_state.current_command,
        Command: parse_state.current_command,
        Output: parse_state.output,
        ParseContext: resolved_context,
        ParseState: parse_state,
        Arg: arg,
        FinalArg: arg,
        Value: Value(result),
        Any: result,
    }
    if option:
        fulfilled_deps[RawOption] = option

    kwargs = fulfill_deps(action_handler, fulfilled_deps).kwargs
    result = action_handler(**kwargs)

    resolved_context.set_result(field_name, result, option, has_value)

    check_deprecated(parse_state, arg, option)


def check_deprecated(
    parse_state: ParseState,
    arg: FinalArg[Any] | FinalCommand[Any],
    option: RawOption | None = None,
) -> None:
    if not arg.deprecated:
        return

    if option:
        kind = "Option"
        name = option.name
    else:
        if isinstance(arg, Command):
            kind = "Command"
            name = arg.real_name()
        else:
            kind = "Argument"
            name = arg.names_str("/")

    message = f"{kind} `{name}` is deprecated"
    if isinstance(arg.deprecated, str):
        message += f": {arg.deprecated}"

    parse_state.output.error(message)


def _warn_deprecated_alias(parse_state: ParseState, alias: Alias) -> None:
    if not alias.deprecated:
        return

    message = f"Command alias `{alias.name}` is deprecated"
    if isinstance(alias.deprecated, str):
        message += f": {alias.deprecated}"

    parse_state.output.error(message)


@dataclasses.dataclass
class Value(Generic[T]):
    value: T


def store_true():
    return True


def store_false():
    return False


def store_count(context: ParseContext, arg: Arg[Any]):
    return context.result.get(cast(str, arg.field_name), 0) + 1


def store_set(value: Value[Any]):
    return value.value


def store_append(context: ParseContext, arg: Arg[Any], value: Value[Any]):
    result = context.result.setdefault(cast(str, arg.field_name), [])
    result.append(value.value)
    return result


def determine_action_handler(action: ArgActionType | None):
    assert action

    if isinstance(action, ArgAction):
        return process_options[action]

    return action


process_options: dict[ArgAction, Callable[..., Any]] = {
    ArgAction.help: HelpAction.from_parse_state,
    ArgAction.version: VersionAction.from_arg,
    ArgAction.completion: CompletionAction.from_value,
    ArgAction.set: store_set,
    ArgAction.store_true: store_true,
    ArgAction.store_false: store_false,
    ArgAction.count: store_count,
    ArgAction.append: store_append,
}



from __future__ import annotations

from dataclasses import dataclass

import pytest
from typing_extensions import Annotated

import cappa
from cappa.output import Exit
from tests.utils import Backend, backends, parse


@dataclass
class Args:
    verbose: Annotated[
        int,
        cappa.Arg(short="-v", action=cappa.ArgAction.count),
        cappa.Arg(long="--verbosity"),
    ] = 0
    meow: Annotated[bool, cappa.Arg(long="--meow/--no-meow")] = False
    bar: Annotated[bool, cappa.Arg(long="--bar/--no-bar")] = False


@backends
def test_help(backend: Backend | None):
    result = parse(Args, "-vvv", "--meow", "--no-bar", backend=backend)
    assert result == Args(verbose=3, meow=True, bar=False)

    result = parse(Args, "--verbosity=4", "--no-meow", "--bar", backend=backend)
    assert result == Args(verbose=4, meow=False, bar=True)

    with pytest.raises(Exit) as e:
        parse(Args, "-vvv", "--verbosity=4", backend=backend)

    if backend is None:
        expected = "Argument '--verbosity' is not allowed with argument '-v'"
    else:
        expected = "argument --verbosity: not allowed with argument -v"
    assert expected.lower() in str(e.value.message).lower()

    with pytest.raises(Exit) as e:
        parse(Args, "--meow", "--no-meow", backend=backend)
    if backend is None:
        expected = "Argument '--no-meow' is not allowed with argument '--meow'"
    else:
        expected = "argument --no-meow: not allowed with argument --meow"
    assert expected.lower() in str(e.value.message).lower()

    args = parse(Args, "--no-bar", "--meow", backend=backend)
    assert args == Args(verbose=0, meow=True, bar=False)



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@dataclass
class Args:
    numbers: Annotated[
        list[int],
        cappa.Arg(
            long=True,
            default=[1, 2, 3],
            num_args=-1,
        ),
    ]


@backends
def test_num_args_unbounded_length_num_args(backend: Backend):
    result = parse(Args, backend=backend)
    assert result == Args([1, 2, 3])

    result = parse(Args, "--numbers", backend=backend)
    assert result == Args([])

    result = parse(Args, "--numbers", "2", backend=backend)
    assert result == Args([2])

    result = parse(Args, "--numbers", "2", "5", backend=backend)
    assert result == Args([2, 5])



from __future__ import annotations

import re
from typing import Any

import pytest
from typing_extensions import Annotated

import cappa
from cappa.output import Exit
from tests.utils import Backend, CapsysOutput, backends, parse


@cappa.command
class Args:
    flag: Annotated[bool, cappa.Arg(long=["--flag", "--no-flag"], help="meow.")] = False


@pytest.mark.help
@backends
def test_default_help(backend: Backend, capsys: Any):
    with pytest.raises(Exit):
        parse(Args, "--help", backend=backend)

    out = CapsysOutput.from_capsys(capsys)
    assert re.match(r".*--no-flag]\s+meow.\s+\(Default", out.stdout, re.DOTALL)


@pytest.mark.help
@backends
def test_default_help_true_default(backend: Backend, capsys: Any):
    """Default text must follow help text regardless of whether default is True or False."""

    @cappa.command
    class TrueDefault:
        flag: Annotated[bool, cappa.Arg(long=["--flag", "--no-flag"], help="meow.")] = (
            True
        )

    with pytest.raises(Exit):
        parse(TrueDefault, "--help", backend=backend)

    out = CapsysOutput.from_capsys(capsys)
    assert re.match(r".*--no-flag]\s+meow.\s+\(Default", out.stdout, re.DOTALL)
    assert "(Default: True)" in out.stdout



from __future__ import annotations

from cappa.completion.types import Completion


def complete_choices(choices: list[str], help: str | None = None):
    def completion(partial: str = ""):
        return [Completion(c, help=help) for c in choices if partial in c]

    return completion



from __future__ import annotations

from dataclasses import dataclass

import pytest
from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


def action():
    raise cappa.Exit()


@backends
def test_single_opt(backend: Backend):
    @dataclass
    class Args:
        kill_switch: Annotated[
            bool, cappa.Arg(long="--kill", action=action, num_args=0)
        ]
        required: str

    with pytest.raises(cappa.Exit) as e:
        parse(Args, "--kill", backend=backend)

    assert e.value.message is None
    assert e.value.code == 0


@backends
def test_num_args_list(backend: Backend):
    @dataclass
    class Args:
        foo: Annotated[
            list[str],
            cappa.Arg(
                short=True,
                default=["", ""],
                required=False,
                num_args=2,
            ),
        ]

    args = parse(Args, backend=backend)
    assert args == Args(["", ""])

    args = parse(Args, "-f", "2", "4", backend=backend)
    assert args == Args(["2", "4"])

    with pytest.raises(cappa.Exit) as e:
        parse(Args, "-f", backend=backend)

    assert e.value.code == 2

    if backend:
        assert str(e.value.message).lower() == "argument -f: expected 2 arguments"
    else:
        assert e.value.message == "Argument '-f' requires 2 values, found 0"


@backends
def test_num_args_optional_value(backend: Backend):
    """NumArgs(required=False) distinguishes absent flag, flag-no-value, flag-with-value."""

    @dataclass
    class Args:
        foo: Annotated[
            int | None,
            cappa.Arg(long=True, num_args=cappa.NumArgs(required=False, default=None)),
        ] = 5

    # Flag not given → class default
    args = parse(Args, backend=backend)
    assert args == Args(foo=5)

    # Flag given with no value → NumArgs.default (None)
    args = parse(Args, "--foo", backend=backend)
    assert args == Args(foo=None)

    # Flag given with value → parsed value
    args = parse(Args, "--foo", "4", backend=backend)
    assert args == Args(foo=4)



import pandas

from cappa.output import Output
from defer_import.main import MeanCommand


def calculate_mean(command: MeanCommand, output: Output):
    if not command.numbers:
        output.output("0.0")
        return

    series = pandas.Series(command.numbers)
    result = float(series.mean())
    output.output(result)



from __future__ import annotations

from dataclasses import dataclass

from tests.utils import parse_completion


def test_arg():
    @dataclass
    class Args:
        value: str

    result = parse_completion(Args, "")
    assert not result


def test_arg_partial():
    @dataclass
    class Args:
        value: str

    result = parse_completion(Args, "m")
    assert result
    assert result == "file"



from __future__ import annotations

from dataclasses import dataclass

import pytest
from typing_extensions import Annotated

import cappa
from cappa.output import Exit
from tests.utils import Backend, backends, parse


@backends
def test_valid(backend: Backend):
    @cappa.command(default_long=True)
    @dataclass
    class Command:
        foo: int = 4
        bar: int = 5

    test = parse(Command, backend=backend)
    assert test == Command(4, 5)

    test = parse(Command, "--foo", "1", "--bar", "2", backend=backend)
    assert test == Command(1, 2)


@backends
def test_override(backend: Backend):
    @cappa.command(default_long=True)
    @dataclass
    class Command:
        foo: int
        far: Annotated[int, cappa.Arg(long="--kar")]

    test = parse(Command, "--foo", "1", "--kar", "2", backend=backend)
    assert test == Command(1, 2)


@backends
def test_invalid(backend: Backend):
    @cappa.command(default_long=True)
    @dataclass
    class Command:
        foo: int
        far: Annotated[int, cappa.Arg(long="--foo")]

    with pytest.raises(Exit) as e:
        parse(Command, backend=backend)
    assert "conflicting option string: --f" in str(e.value.message).lower()



from __future__ import annotations

import importlib
import importlib.util
import io
import logging
from typing import TYPE_CHECKING, Any, ClassVar, Sequence

from rich.console import Console

import cappa
from cappa.command import FinalCommand
from cappa.default import Default
from cappa.help import ArgGroup, HelpFormatter
from cappa.output import Displayable, theme
from cappa.type_view import Empty

try:
    importlib.util.find_spec("docutils")
except ImportError:  # pragma: no cover
    raise ImportError(
        "The ext.docutils package can only be used if 'docutils' is installed."
    )

from docutils import nodes
from docutils.parsers.rst import Directive, directives

if TYPE_CHECKING:
    from sphinx.application import Sphinx

log = logging.getLogger(__name__)

FONT_FAMILY = (
    """font-family:Menlo,"DejaVu Sans Mono",consolas,"Courier New",monospace"""
)


def setup(app: Sphinx) -> dict[str, Any]:  # pragma: no cover
    app.add_directive("cappa", CappaDirective)
    return {"parallel_read_safe": True}


class CappaDirective(Directive):
    has_content = False
    required_arguments = 1
    option_spec: ClassVar[dict[str, Any] | None] = {
        "style": directives.unchanged,
        "terminal-width": directives.unchanged,
    }

    def run(self) -> Sequence[nodes.Node]:
        path = self.arguments[0]
        style = self.options.get("style", "terminal")
        terminal_width = int(self.options.get("terminal-width", "0"))

        module_path, item_name = path.rsplit(".", 1)

        module = importlib.import_module(module_path)
        item = getattr(module, item_name)

        command: FinalCommand[Any] = cappa.collect(item)
        if style == "terminal":
            return render_to_terminal(command, terminal_width)

        if style == "native":
            return render_to_docutils(command, self.state.document)

        raise ValueError(f"Unrecognized style {style}")


def render_to_terminal(command: cappa.FinalCommand[Any], terminal_width: int):
    raw_help: list[Displayable] = HelpFormatter.default.long(
        command, command.real_name()
    )

    line_wrap = ""
    if terminal_width == 0:
        line_wrap = "white-space: pre-wrap;"
        terminal_width = 99999

    console = Console(
        record=True,
        theme=theme,
        file=io.StringIO(),
        soft_wrap=True,
        width=terminal_width,
    )
    console.print(*raw_help)

    style = " ".join([line_wrap, FONT_FAMILY])

    code_format = f'<pre style="{style}"><code>{{code}}</code></pre>'
    html = console.export_html(inline_styles=True, code_format=code_format)
    return [nodes.raw(text=html, format="html")]


def render_to_docutils(command: cappa.FinalCommand[Any], document: nodes.document):
    section = nodes.section()
    document.note_implicit_target(section)

    section += nodes.title(text=command.real_name())

    command_content = nodes.paragraph()
    section += command_content

    description: list[str] = []
    if command.help:
        description.append(command.help)
    if command.description:
        description.append(command.description)

    section += nodes.paragraph(text="\n\n".join(description))

    for group in ArgGroup.collect(command):
        # Collect all Args and Subcommands from field_groups
        command_options: list[cappa.FinalArg[Any]] = []
        command_subcommands: list[cappa.FinalSubcommand] = []
        for field_group in group.field_groups:
            if field_group.args:
                command_options.extend(field_group.args)
            else:
                assert field_group.subcommand
                command_subcommands.append(field_group.subcommand)

        if command_options:
            section += nodes.subtitle(text=group.name)

            option = nodes.bullet_list()
            section += option

            for arg in command_options:
                option_content = nodes.list_item()
                option += option_content

                names = arg.names()
                if names:
                    for name in names:
                        option_content += nodes.literal(text=name)
                else:
                    name = arg.field_name
                    value_name = arg.value_name

                    name += f" {value_name.upper()}"
                    option_content += nodes.literal(text=name)

                assert isinstance(arg.default, Default)
                if arg.default.default not in (Empty, None):
                    default = str(arg.default)

                    option_content += nodes.inline(text=" (")
                    option_content += nodes.literal(text=default)
                    option_content += nodes.inline(text=")")

                if arg.help:
                    option_content += nodes.inline(text=f": {arg.help}")

        if command_subcommands:
            for subcmd in command_subcommands:
                for o in subcmd.options.values():
                    section += render_to_docutils(o, document)

    return [section]



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, TypedDict

from typing_extensions import Annotated

import cappa
from cappa.default import ValueFrom
from cappa.state import State
from tests.utils import Backend, backends, invoke, parse


class Increment(TypedDict):
    value: int


def woah(state: State[Increment]):
    state.state["value"] = state.state.get("value", 0) + 1
    return state.state["value"]


@dataclass
class Command:
    foo: Annotated[int, cappa.Arg(default=ValueFrom(woah))] = 0
    bar: Annotated[int, cappa.Arg(default=ValueFrom(woah))] = 0

    def __call__(self, state: State[Increment]):
        return state.state


@backends
def test_shared_state(backend: Backend):
    result = parse(Command, "6", "7", backend=backend)
    assert result == Command(6, 7)

    result = parse(Command, backend=backend)
    assert result == Command(1, 2)


@backends
def test_dep_state(backend: Backend):
    result = invoke(Command, "6", "7", backend=backend)
    assert result == {}

    result = invoke(Command, backend=backend)
    assert result == {"value": 2}


def parse_val(value: str, state: State[dict[str, Any]]):
    state.set("foo", int(value))
    return state.get("foo")


@backends
def test_parse_state(backend: Backend):
    state: State[dict[str, Any]] = State()

    @dataclass
    class Command:
        foo: Annotated[int, cappa.Arg(parse=parse_val)] = 0

    result = parse(Command, backend=backend, state=state)
    assert result == Command(0)
    assert state.get("foo") is None

    result = parse(Command, "6", backend=backend, state=state)
    assert result == Command(6)
    assert state.get("foo") == 6



from __future__ import annotations

from dataclasses import dataclass
from typing import Union

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_collect_post_dash_args(backend: Backend):
    @dataclass
    class Args:
        foo: str
        raw: Annotated[Union[list[str], None], cappa.Arg(num_args=-1)] = None

    test = parse(Args, "foovalue", "--", "--raw", "value", backend=backend)
    assert test.foo == "foovalue"
    assert test.raw == ["--raw", "value"]



from __future__ import annotations

from dataclasses import dataclass
from typing import Literal

import pytest

import cappa
from tests.utils import Backend, backends, parse


@dataclass
class ArgTest:
    name: Literal["one", "two", "three", 4]


@backends
def test_literal(backend: Backend):
    test = parse(ArgTest, "two", backend=backend)
    assert test.name == "two"

    test = parse(ArgTest, "4", backend=backend)
    assert test.name == 4

    with pytest.raises(cappa.Exit) as e:
        parse(ArgTest, "thename", backend=backend)

    message = str(e.value.message).lower()
    assert "invalid choice: 'thename' (choose from 'one', 'two', 'three', 4)" in message



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import pytest

import cappa
from tests.utils import Backend, backends, parse


@dataclass
class One: ...


@cappa.command(hidden=True)
@dataclass
class Two: ...


@dataclass
class Three: ...


@dataclass
class Command:
    sub: cappa.Subcommands[One | Two | Three]


@backends
def test_has_possible_values(capsys: Any, backend: Backend):
    with pytest.raises(cappa.Exit):
        parse(Command, "--help", backend=backend)

    out = capsys.readouterr().out
    assert "one" in out.lower()
    assert "two" not in out.lower()
    assert "three" in out.lower()

    cmd = parse(Command, "two", backend=backend)
    assert isinstance(cmd.sub, Two)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import pytest

import cappa
from tests.utils import parse


@dataclass
class Args:
    arg: str


def test_option_after_bad_arg(capsys: Any):
    with pytest.raises(cappa.Exit):
        parse(Args, "arg1", "arg2", "--help")
    err = capsys.readouterr().err
    assert "Unrecognized arguments: arg2" in err



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_count_action(backend: Backend):
    @dataclass
    class ArgTest:
        arg: Annotated[int, cappa.Arg(short=True, action=cappa.ArgAction.count)]

    result = parse(ArgTest, "-a", backend=backend)
    assert result.arg == 1

    result = parse(ArgTest, "-aaa", backend=backend)
    assert result.arg == 3


@backends
def test_count_option(backend: Backend):
    @dataclass
    class ArgTest:
        arg: Annotated[int, cappa.Arg(short=True, count=True)]

    result = parse(ArgTest, "-a", backend=backend)
    assert result.arg == 1

    result = parse(ArgTest, "-aaa", backend=backend)
    assert result.arg == 3



from __future__ import annotations

from pathlib import Path
from typing import Any


def successful_test_run(
    pytester: Any, *, count: int | None = None, skipped_count: int = 0
):
    pytester.copy_example()
    result = pytester.runpytest_inprocess(
        "-vv", "--log-cli-level=WARNING", "--log-level=WARNING", "conftest.py"
    )
    result.assert_outcomes(passed=count, skipped=skipped_count, failed=0)


def test_kitchen_sink(pytester: Any):
    successful_test_run(pytester, count=1)


def test_docutils(pytester: Any):
    build_path: Path = pytester.path / "_build"
    path = build_path / "index.html"
    path.unlink(missing_ok=True)

    pytester.copy_example(name="docutils")
    pytester.run("sphinx-build", "-W", "--jobs=2", pytester.path, build_path)

    output = path.read_text()
    assert "Usage: foo" in output



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Union

import pytest
from type_lens.type_view import TypeView
from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


def split(value: str):
    a, b = value.split(",", 1)
    return int(a), int(b)


def parse_int(val: str) -> int:
    try:
        return int(val)
    except Exception:
        raise cappa.Exit("no go amego", code=2)


@backends
def test_explicit_parse_function(backend: Backend):
    @dataclass
    class ArgTest:
        numbers: Annotated[int, cappa.Arg(parse=int)]

    test = parse(ArgTest, "1", backend=backend)
    assert test.numbers == 1


@backends
def test_not_typeable_parse_function(backend: Backend):
    @dataclass
    class ArgTest:
        numbers: Annotated[tuple[int, int], cappa.Arg(parse=split)]

    test = parse(ArgTest, "1,3", backend=backend)
    assert test.numbers == (1, 3)


@backends
def test_parse_failure(backend: Backend):
    """Optional annotations should take precedence over explicit parse."""

    @dataclass
    class ArgTest:
        numbers: Annotated[int, cappa.Arg(parse=parse_int)]

    with pytest.raises(cappa.Exit) as e:
        parse(ArgTest, "one", backend=backend)

    assert e.value.code == 2
    assert e.value.message == "no go amego"


@backends
def test_parse_optional(backend: Backend):
    """Optionals are handled by the given/inferred parse method."""

    @dataclass
    class ArgTest:
        numbers: Annotated[Union[int, None], cappa.Arg(parse=float)] = None

    result = parse(ArgTest, backend=backend)
    assert result == ArgTest()

    with pytest.raises(cappa.Exit) as e:
        parse(ArgTest, "one", backend=backend)

    assert e.value.code == 2
    assert (
        e.value.message
        == "Invalid value for 'numbers': could not convert string to float: 'one'"
    )


def parse_test_parse_returns_none(value: str) -> bool | None:
    if int(value) % 2:
        return True
    return None


@backends
def test_parse_returns_none(backend: Backend):
    """Assert that a parser which returns None succeeds ."""

    @dataclass
    class ArgTest:
        numbers: Annotated[int | None, cappa.Arg(parse=parse_test_parse_returns_none)]

    result = parse(ArgTest, "1", backend=backend)
    assert result == ArgTest(True)

    result = parse(ArgTest, "2", backend=backend)
    assert result == ArgTest(None)


def parser_test_typed_parse(value: str):
    return float(value)


@backends
def test_typed_parse(backend: Backend):
    """A parse function with a typed argument."""

    @dataclass
    class ArgTest:
        num: Annotated[float, cappa.Arg(parse=parser_test_typed_parse)]

    result = parse(ArgTest, "4.1", backend=backend)
    assert result == ArgTest(num=4.1)


def parser_test_type_aware_parse(value: str, type_view: TypeView[Any]) -> str:
    return str(type_view)


@backends
def test_type_aware_parse(backend: Backend):
    """A parse function that receives type information."""

    @dataclass
    class ArgTest:
        num: Annotated[str, cappa.Arg(parse=parser_test_type_aware_parse)]

    result = parse(ArgTest, "4.1", backend=backend)
    assert result == ArgTest(num="TypeView(str)")


def parse_exit(_: Any) -> Any:
    raise cappa.Exit("asdf")


@backends
def test_parse_exit(backend: Backend):
    @dataclass
    class ArgTest:
        num: Annotated[str, cappa.Arg(parse=parse_exit)]

    with pytest.raises(cappa.Exit):
        parse(ArgTest, "4.1", backend=backend)


def min0(value: int) -> int:
    if not value > 0:
        raise ValueError(f"{value} > 0")
    return value


@backends
def test_default_parse(backend: Backend):
    @dataclass
    class ArgTest:
        num: Annotated[int, cappa.Arg(parse=[cappa.default_parse, min0])]

    result = parse(ArgTest, "4", backend=backend)
    assert result.num == 4

    with pytest.raises(cappa.Exit) as e:
        parse(ArgTest, "--", "-4", backend=backend)
    assert e.value.message == "Invalid value for 'num': -4 > 0"



from __future__ import annotations

import sys
from dataclasses import dataclass
from typing import Any

import pytest

import cappa
from tests.utils import parse


@cappa.command(deprecated=True)
@dataclass
class Sub:
    pass


@dataclass
class ArgTest:
    sub: cappa.Subcommands[Sub | None] = None


def test_native_backend(capsys: Any):
    """Note, argparse only support deprecated at or above 3.13."""
    result = parse(ArgTest)
    assert result.sub is None
    err = capsys.readouterr().err
    assert err == ""

    result = parse(ArgTest, "sub")
    assert result.sub == Sub()
    err = capsys.readouterr().err
    assert err == "Error: Command `sub` is deprecated\n"


@pytest.mark.skipif(sys.version_info >= (3, 13), reason="requires python3.13 or higher")
def test_argparse_le_313(capsys: Any):
    """Below 3.13, this option has no effect."""
    result = parse(ArgTest, backend=cappa.argparse.backend)
    assert result.sub is None
    err = capsys.readouterr().err
    assert err == ""

    result = parse(ArgTest, "sub", backend=cappa.argparse.backend)
    assert result.sub == Sub()
    err = capsys.readouterr().err
    assert err == ""


@pytest.mark.skipif(
    sys.version_info < (3, 13), reason="Below 3.13, the behavior is different"
)
def test_argparse_ge_313(capsys: Any):
    result = parse(ArgTest, backend=cappa.argparse.backend)
    assert result.sub is None
    err = capsys.readouterr().err
    assert err == ""

    result = parse(ArgTest, "sub", backend=cappa.argparse.backend)
    assert result.sub == Sub()
    err = capsys.readouterr().err.replace("arg-test: ", "").strip()
    assert err == "warning: command 'sub' is deprecated"



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_short_missing_dash(backend: Backend):
    @dataclass
    class ArgTest:
        number: Annotated[int, cappa.Arg(short="n")]

    result = parse(ArgTest, "-n", "4", backend=backend)
    assert result.number == 4


@backends
def test_multiple_shorts(backend: Backend):
    @dataclass
    class ArgTest:
        number: Annotated[int, cappa.Arg(short=["n", "o"])]

    result = parse(ArgTest, "-n", "4", backend=backend)
    assert result.number == 4

    result = parse(ArgTest, "-o", "4", backend=backend)
    assert result.number == 4


@backends
def test_long_missing_dash(backend: Backend):
    @dataclass
    class ArgTest:
        number: Annotated[int, cappa.Arg(long="nu")]

    result = parse(ArgTest, "--nu", "4", backend=backend)
    assert result.number == 4


@backends
def test_multiple_longs(backend: Backend):
    @dataclass
    class ArgTest:
        number: Annotated[int, cappa.Arg(long=["so", "long"])]

    result = parse(ArgTest, "--so", "4", backend=backend)
    assert result.number == 4

    result = parse(ArgTest, "--long", "4", backend=backend)
    assert result.number == 4



from __future__ import annotations

import argparse
import sys
from typing import TYPE_CHECKING, Any, Callable, Hashable, TypeVar, cast

from cappa.arg import Arg, ArgAction, FinalArg, NumArgs
from cappa.command import Alias, Command, FinalCommand
from cappa.help import ArgGroup
from cappa.invoke.base import fulfill_deps
from cappa.output import Exit, HelpExit, Output
from cappa.parser import RawOption, Value
from cappa.subcommand import FinalSubcommand

if TYPE_CHECKING:
    from _typeshed import SupportsWrite

if sys.version_info < (3, 9):  # pragma: no cover
    # Backport https://github.com/python/cpython/pull/3680
    original_get_action_name = argparse._get_action_name  # pyright: ignore

    def _get_action_name(
        argument: argparse.Action | None,
    ) -> str | None:  # pragma: no cover
        name = original_get_action_name(argument)

        assert argument
        if name is None and argument.choices:
            return "{" + ",".join(argument.choices) + "}"

        return name

    argparse._get_action_name = _get_action_name  # pyright: ignore


T = TypeVar("T")


# Work around argparse's lack of `metavar` support on various built-in actions.
class _HelpAction(argparse._HelpAction):  # pyright: ignore
    def __init__(self, metavar: str | None = None, **kwargs: Any):
        self.metavar = metavar
        super().__init__(**kwargs)


class _VersionAction(argparse._VersionAction):  # pyright: ignore
    def __init__(self, metavar: str | None = None, **kwargs: Any):
        self.metavar = metavar
        super().__init__(**kwargs)


class _StoreTrueAction(argparse._StoreTrueAction):  # pyright: ignore
    def __init__(self, metavar: str | None = None, **kwargs: Any):
        self.metavar = metavar
        super().__init__(**kwargs)


class _StoreFalseAction(argparse._StoreFalseAction):  # pyright: ignore
    def __init__(self, metavar: str | None = None, **kwargs: Any):
        self.metavar = metavar
        super().__init__(**kwargs)


class _CountAction(argparse._CountAction):  # pyright: ignore
    def __init__(self, metavar: str | None = None, **kwargs: Any):
        self.metavar = metavar
        super().__init__(**kwargs)


class ArgumentParser(argparse.ArgumentParser):
    def __init__(
        self, *args: Any, command: FinalCommand[Any], output: Output, **kwargs: Any
    ):
        super().__init__(*args, **kwargs)
        self.command = command
        self.output = output

        self.register("action", "store_true", _StoreTrueAction)
        self.register("action", "store_false", _StoreFalseAction)
        self.register("action", "help", _HelpAction)
        self.register("action", "version", _VersionAction)
        self.register("action", "count", _CountAction)

    def error(self, message: str):
        # Avoids argparse's error prefixing code, deferring it to Output
        self.exit(2, message)

    def exit(self, status: int = 0, message: str | None = None):
        if message:
            message = message.capitalize()

        raise Exit(message, code=status, prog=self.prog)

    def print_help(self, file: SupportsWrite[str] | None = None):
        raise HelpExit(self.command.help_formatter.long(self.command, self.prog))


def custom_action(arg: Arg[Any], action: Callable[..., Any]):
    class CustomAction(argparse.Action):
        def __call__(  # type: ignore
            self,
            parser: ArgumentParser,
            namespace: argparse.Namespace,
            values: Any,
            option_string: str | None = None,
        ):
            # XXX: This should ideally be able to inject parser state, but here, we dont
            #      have access to the same state as the native parser.
            fulfilled_deps: dict[Hashable, Any] = {
                Output: parser.output,
                Value: Value(values),
                Command: namespace.__command__,
                Arg: arg,
            }
            if option_string:
                fulfilled_deps[RawOption] = RawOption.from_str(option_string)

            deps = fulfill_deps(action, fulfilled_deps)
            result = action(**deps.kwargs)
            setattr(namespace, self.dest, result)

    return CustomAction


class Nestedspace(argparse.Namespace):
    """Write each . separated section as a nested `Nestedspace` instance.

    By default, argparse write everything to a flat namespace so there's no
    obvious way to distinguish between multiple unrelated subcommands once
    once has been chosen.
    """

    def __setattr__(self, name: str, value: Any):
        if "." in name:
            group, name = name.split(".", 1)
            ns = getattr(self, group, Nestedspace())
            setattr(ns, name, value)
            self.__dict__[group] = ns
        else:
            self.__dict__[name] = value


def backend(
    command: FinalCommand[T],
    argv: list[str],
    output: Output,
    prog: str,
    provide_completions: bool = False,
) -> tuple[Any, Command[T], dict[str, Any]]:
    parser = create_parser(command, output=output, prog=prog)

    try:
        version = next(
            a
            for a in command.arguments
            if isinstance(a, Arg) and a.action is ArgAction.version
        )
        parser.version = version.value_name  # type: ignore
    except StopIteration:
        pass

    ns = Nestedspace()

    try:
        result_namespace = parser.parse_args(argv, ns)
    except argparse.ArgumentError as e:
        raise Exit(str(e), code=2, prog=prog)

    result = to_dict(result_namespace)
    command = result.pop("__command__")

    return parser, command, result


def create_parser(
    command: FinalCommand[Any], output: Output, prog: str
) -> argparse.ArgumentParser:
    kwargs: dict[str, Any] = {}
    if sys.version_info >= (3, 9):  # pragma: no cover
        kwargs["exit_on_error"] = False

    parser = ArgumentParser(
        command=command,
        output=output,
        prog=prog,
        description=join_help(command.help, command.description),
        allow_abbrev=False,
        add_help=False,
        **kwargs,
    )
    parser.set_defaults(__command__=command)

    add_arguments(parser, command, output=output)
    return parser


def add_arguments(
    parser: ArgumentParser,
    command: FinalCommand[Any],
    output: Output,
    dest_prefix: str = "",
):
    exclusive_groups: dict[str, Any] = {}

    arg_groups = ArgGroup.collect(command, include_hidden=True)
    for group in arg_groups:
        argparse_group = parser.add_argument_group(title=group.name)

        for field_group in group.field_groups:
            if field_group.args:
                for arg in field_group.args:
                    target_group: Any = argparse_group

                    if arg.group.exclusive:
                        target_group = exclusive_groups.get(arg.group.id)
                        if target_group is None:
                            target_group = exclusive_groups[arg.group.id] = (
                                argparse_group.add_mutually_exclusive_group()
                            )

                    add_argument(parser, target_group, arg, dest_prefix=dest_prefix)
            else:
                assert field_group.subcommand
                add_subcommands(
                    parser,
                    group.name,
                    field_group.subcommand,
                    output=output,
                    dest_prefix=dest_prefix,
                )


def add_argument(
    parser: ArgumentParser,  # pyright: ignore
    subparser: argparse.ArgumentParser | argparse._ArgumentGroup,  # pyright: ignore
    arg: FinalArg[Any],
    dest_prefix: str = "",
    **extra_kwargs: Any,
):
    if arg.propagate:
        raise RuntimeError("The argparse backend does not support the `Arg.propagate`.")

    names: list[str] = []
    if arg.short:
        names.extend(arg.short)

    if arg.long:
        names.extend(arg.long)

    is_positional = not names

    num_args = backend_num_args(arg.num_args, arg.required)

    kwargs: dict[str, Any] = {
        "dest": dest_prefix + arg.field_name,
        "help": arg.help,
        "metavar": arg.value_name,
        "action": get_action(arg),
        "default": argparse.SUPPRESS,
    }

    if not is_positional and arg.required and arg.num_args.n >= 0:
        kwargs["required"] = arg.required

    is_optional_value = not is_positional and not arg.num_args.required

    if num_args is not None and not ArgAction.is_non_value_consuming(arg.action):
        kwargs["nargs"] = num_args
    elif is_optional_value:
        kwargs["nargs"] = "?"
        kwargs["const"] = arg.num_args.default
    elif is_positional and not arg.required:
        kwargs["nargs"] = "?"

    deprecated_kwarg = add_deprecated_kwarg(arg)
    kwargs.update(deprecated_kwarg)
    kwargs.update(extra_kwargs)

    try:
        subparser.add_argument(*names, **kwargs)
    except argparse.ArgumentError as e:
        raise Exit(str(e), code=2, prog=parser.prog)


def add_subcommands(
    parser: argparse.ArgumentParser,
    group: str,
    subcommands: FinalSubcommand,
    output: Output,
    dest_prefix: str = "",
):
    subcommand_dest = subcommands.field_name
    visible_metavar = "{" + ",".join(subcommands.names()) + "}"
    subparsers = parser.add_subparsers(
        title=group,
        required=subcommands.required,
        parser_class=ArgumentParser,
        metavar=visible_metavar,
    )

    for name, subcommand in subcommands.options.items():
        deprecated_kwarg = add_deprecated_kwarg(subcommand)

        nested_dest_prefix = f"{dest_prefix}{subcommand_dest}."
        non_deprecated_visible_aliases = [
            a.name
            for a in subcommand.resolved_aliases()
            if not a.hidden and not a.deprecated
        ]
        subparser = subparsers.add_parser(
            name=subcommand.real_name(),
            aliases=non_deprecated_visible_aliases,
            help=subcommand.help,
            description=subcommand.description,
            formatter_class=parser.formatter_class,
            add_help=False,
            command=subcommand,
            output=output,
            prog=f"{parser.prog} {subcommand.real_name()}",
            **deprecated_kwarg,
        )
        subparser.set_defaults(
            __command__=subcommand, **{nested_dest_prefix + "__name__": name}
        )

        # Aliases not registered via `aliases=` (hidden, or deprecated) get spliced
        # into the internal name map. Deprecated aliases get a wrapper that emits
        # the warning at dispatch time, then delegates parsing to the real subparser.
        for alias in subcommand.resolved_aliases():
            if alias.name in non_deprecated_visible_aliases:
                continue
            if alias.deprecated:
                subparsers._name_parser_map[alias.name] = _DeprecatedAliasParser(
                    subparser, alias, output
                )
            else:
                subparsers._name_parser_map[alias.name] = subparser

        add_arguments(
            subparser,
            subcommand,
            output=output,
            dest_prefix=nested_dest_prefix,
        )


def backend_num_args(num_args: NumArgs, required: bool) -> int | str | None:
    if num_args.n == 1:
        return None

    if num_args.n == -1:
        if required:
            return "+"
        return "*"

    return num_args.n


class _DeprecatedAliasParser(ArgumentParser):
    """Stand-in subparser used when a user invokes a deprecated alias.

    argparse dispatches into here via `_name_parser_map[alias_name]`. We emit
    the deprecation warning, then delegate parsing to the real subparser so the
    canonical `__name__` and arguments are populated as usual.
    """

    def __init__(
        self, real_parser: ArgumentParser, alias: Alias, output: Output
    ) -> None:
        super().__init__(
            command=real_parser.command,
            output=output,
            add_help=False,
            allow_abbrev=False,
        )
        self._real_parser = real_parser
        self._alias = alias

    def parse_known_args(  # type: ignore[override]
        self,
        args: Any = None,
        namespace: Any = None,
    ) -> tuple[argparse.Namespace, list[str]]:
        message = f"Command alias `{self._alias.name}` is deprecated"
        if isinstance(self._alias.deprecated, str):
            message += f": {self._alias.deprecated}"
        self.output.error(message)
        return self._real_parser.parse_known_args(args, namespace)


def to_dict(value: argparse.Namespace):
    result: dict[str, Any] = {}
    for k, v in value.__dict__.items():
        if isinstance(v, argparse.Namespace):
            v = to_dict(v)
        result[k] = v

    return result


def join_help(*segments: str | None):
    return " ".join([s for s in segments if s])


def get_action(arg: Arg[Any]) -> argparse.Action | type[argparse.Action] | str:
    action = arg.action
    if isinstance(action, ArgAction):
        return action.value

    action = cast(Callable[..., Any], action)
    return custom_action(arg, action)


def add_deprecated_kwarg(arg: Arg[Any] | Command[Any]) -> dict[str, Any]:
    if sys.version_info < (3, 13) or not arg.deprecated:
        return {}

    return {"deprecated": arg.deprecated}  # pragma: no cover



from __future__ import annotations

import sys
from typing import Optional

import pytest
from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse

if sys.version_info >= (3, 9):
    try:
        import msgspec

        HAVE_MSGSPEC = True

        class PydanticCommand(msgspec.Struct):
            name: str
            foo: Annotated[int, cappa.Arg(short=True)]

        class OptSub(msgspec.Struct):
            name: Optional[str] = None

        class OptionalSubcommand(msgspec.Struct):
            sub: cappa.Subcommands[Optional[OptSub]] = None

    except ImportError:
        HAVE_MSGSPEC = False  # pyright: ignore[reportConstantRedefinition]

    @pytest.mark.skipif(not HAVE_MSGSPEC, reason="msgspec not installed")
    @backends
    def test_base_model(backend: Backend):
        result = parse(PydanticCommand, "meow", "-f", "4", backend=backend)
        assert result == PydanticCommand(name="meow", foo=4)

    @pytest.mark.skipif(not HAVE_MSGSPEC, reason="msgspec not installed")
    @backends
    def test_optional_subcommand(backend: Backend):
        result = parse(OptionalSubcommand, backend=backend)
        assert result == OptionalSubcommand(sub=None)

        result = parse(OptionalSubcommand, "opt-sub", backend=backend)
        assert result == OptionalSubcommand(sub=OptSub(name=None))

        result = parse(OptionalSubcommand, "opt-sub", "foo", backend=backend)
        assert result == OptionalSubcommand(sub=OptSub(name="foo"))



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import pytest

import cappa
from tests.utils import parse


@dataclass
class SubTwo: ...


@dataclass
class SubOne:
    sub_two: cappa.Subcommands[SubTwo]


@dataclass
class Args:
    sub_one: cappa.Subcommands[SubOne]


def test_all_help_options_share_defined_help_argument(capsys: Any):
    help: cappa.Arg[Any] = cappa.Arg(help="this is custom text", long=True)

    with pytest.raises(cappa.HelpExit):
        parse(Args, "--help", help=help)
    out = capsys.readouterr().out
    assert "this is custom text" in out

    with pytest.raises(cappa.HelpExit):
        parse(Args, "sub-one", "--help", help=help)
    out = capsys.readouterr().out
    assert "this is custom text" in out

    with pytest.raises(cappa.HelpExit):
        parse(Args, "sub-one", "sub-two", "--help", help=help)
    out = capsys.readouterr().out
    assert "this is custom text" in out



from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@dataclass
class Args:
    arg: Annotated[int, cappa.Arg(short="-a", long=True)]
    foo: Annotated[bool, cappa.Arg(short="-f", long=True)] = False
    bar: Annotated[bool, cappa.Arg(short="-b", long=True)] = False


@backends
def test_single_option_value_no_space(backend: Backend):
    args = parse(Args, "-a0", backend=backend)
    assert args == Args(arg=0, foo=False, bar=False)


@backends
def test_single_option_value_equals(backend: Backend):
    args = parse(Args, "-a=0", backend=backend)
    assert args == Args(arg=0, foo=False, bar=False)


@backends
def test_multiple_option_no_space(backend: Backend):
    args = parse(Args, "-fba0", backend=backend)
    assert args == Args(arg=0, foo=True, bar=True)


@backends
def test_distinguish_short_args_with_args_from_without(backend: Backend):
    # Only short args which consume no additional values (i.e. flags) should
    # be captured as concatenated short args. With arg which consumes a value,
    # subsequent characters should be given as part of the "arg" portion of
    # the value
    @dataclass
    class Args:
        arg: Annotated[str, cappa.Arg(short="-a")]
        foo: Annotated[bool, cappa.Arg(short="-f")] = False

    args = parse(Args, "-fahello world", backend=backend)
    assert args == Args(arg="hello world", foo=True)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, invoke


def level_three():
    return {"level_three": True}


def level_two(level_three: Annotated[dict[str, Any], cappa.Dep(level_three)]):
    return {"level_two": {**level_three}}


def level_one(level_two: Annotated[dict[str, Any], cappa.Dep(level_two)]):
    return {"level_one": {**level_two}}


def command(levels: Annotated[dict[str, Any], cappa.Dep(level_one)]):
    return levels


@cappa.command(invoke=command)
@dataclass
class Command: ...


@backends
def test_invoke_top_level_command(backend: Backend):
    result = invoke(Command, backend=backend)
    assert result == {"level_one": {"level_two": {"level_three": True}}}



from __future__ import annotations

import ast
import inspect
import textwrap
import typing
from dataclasses import dataclass
from inspect import cleandoc
from types import ModuleType

from typing_extensions import Self

try:
    import docstring_parser as _docstring_parser

    docstring_parser: ModuleType | None = _docstring_parser
except ImportError:  # pragma: no cover
    docstring_parser = None


@dataclass
class ClassHelpText:
    summary: str | None
    body: str | None
    args: dict[str, str]

    @classmethod
    def collect(cls, command: type) -> Self:
        args: dict[str, str] = {}

        doc = get_doc(command)
        if docstring_parser:
            parsed_help = docstring_parser.parse(doc)
            for param in parsed_help.params:
                args[param.arg_name] = param.description
            summary = parsed_help.short_description
            body = parsed_help.long_description
        else:
            doc = inspect.cleandoc(doc).split("\n", 1)
            if len(doc) == 1:
                summary = doc[0]
                body = ""
            else:
                summary, body = doc
                body = body.strip()

        try:
            ast_args = get_attribute_docstrings(command)
        except Exception:
            ast_args = {}

        args.update(ast_args)

        return cls(summary=summary, body=body, args=args)


def get_doc(cls: type):
    """Lifted from dataclasses source."""
    doc = cls.__doc__ or ""

    # Dataclasses will set the doc attribute to the below value if there was no
    # explicit docstring. This is just annoying for us, so we treat that as though
    # there wasn't one.
    try:
        # In some cases fetching a signature is not possible.
        # But, we surely should not fail in this case.
        text_sig = str(inspect.signature(cls)).replace(" -> None", "")
    except (TypeError, ValueError):  # pragma: no cover
        text_sig = ""

    dataclasses_docstring = cls.__name__ + text_sig

    if doc == dataclasses_docstring:
        return ""
    return doc


def get_attribute_docstrings(command: type) -> dict[str, str]:
    result: dict[str, str] = {}

    raw_source = inspect.getsource(command)
    source = textwrap.dedent(raw_source)
    module = ast.parse(source)

    cls_node = module.body[-1]
    assert isinstance(cls_node, ast.ClassDef)

    last_assignment: ast.AnnAssign | None = None
    for node in cls_node.body:
        if isinstance(node, ast.Expr):
            if last_assignment:
                name = typing.cast(ast.Name, last_assignment.target).id
                value = typing.cast(ast.Constant, node.value).value
                result[name] = cleandoc(str(value))
                continue

        last_assignment = node if isinstance(node, ast.AnnAssign) else None

    return result



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Union

import pytest
from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_default(backend: Backend):
    @dataclass
    class ArgTest:
        default: int

    with pytest.raises(cappa.Exit) as e:
        parse(ArgTest, "foo", backend=backend)

    assert e.value.code == 2
    assert (
        e.value.message
        == "Invalid value for 'default': invalid literal for int() with base 10: 'foo'"
    )


@backends
def test_other_exception_types(backend: Backend):
    @dataclass
    class ArgTest:
        path: Annotated[Union[Path, None], cappa.Arg(long=True)] = None

    result = parse(ArgTest, backend=backend)
    assert result.path is None

    result = parse(ArgTest, "--path", "asdf", backend=backend)
    assert result.path == Path("asdf")



from __future__ import annotations

from dataclasses import dataclass, replace
from typing import Any, Generic, TextIO, TypeVar

from type_lens import TypeView
from typing_extensions import Annotated

from cappa.arg import FinalArg
from cappa.default import Default
from cappa.invoke.types import Resolved
from cappa.output import Output
from cappa.state import State


@dataclass(frozen=True)
class Destructure:
    """Collection for destructuring settings."""

    @classmethod
    def collect(
        cls,
        field_name: str,
        default: Default,
        destructure: Destructure | bool | None,
        type_view: TypeView[Any],
        default_short: bool = False,
        default_long: bool = False,
    ) -> FinalDestructure[Any] | None:
        if not destructure:
            return None

        if destructure is True:
            destructure = Destructure()

        annotation = type_view.strip_optional().annotation
        if not isinstance(annotation, type):
            raise ValueError(
                "Destructured arguments currently only support singular concrete types."
            )

        inner: Command[Any] = Command.get(annotation)  # pyright: ignore
        inner = replace(
            inner,
            default_short=inner.default_short or default_short,
            default_long=inner.default_long or default_long,
        )
        command: FinalCommand[Any] = inner.collect()
        return FinalDestructure(
            field_name=field_name,
            command=command,
            is_optional=type_view.is_optional,
            default=default,
        )


T = TypeVar("T")
Destructured = Annotated[T, Destructure()]


@dataclass(frozen=True)
class FinalDestructure(Destructure, Generic[T]):
    """Post-normalization representation of a destructured argument.

    This is produced alongside FinalArg/FinalSubcommand in a `Command`'s set of arguments.
    The parser and mapper work together to produce the destructured value as a nested mapping
    that the contained virtual `command` can understand, and then the result is mapped to
    the final value of the field.
    """

    field_name: str
    command: FinalCommand[T]
    is_optional: bool
    default: Default

    def map_result(
        self,
        prog: str,
        parsed_args: dict[str, Any],
        output: Output,
        state: State[Any] | None = None,
        input: TextIO | None = None,
    ) -> Resolved[T]:
        if self.is_optional and not parsed_args:
            _, value = self.default(state=state, input=input)
            return Resolved(lambda: value)

        return self.command.map_result(
            self.command, prog, parsed_args, output, state, input
        )[0]

    def explode_args(self) -> list[FinalArg[Any] | FinalDestructure[Any]]:
        virtual_args = self.command.arguments

        result: list[FinalArg[Any] | FinalDestructure[Any]] = [self]
        for virtual_arg in virtual_args:
            if not isinstance(virtual_arg, FinalArg):
                raise ValueError(
                    "Only `Arg` is supported in the context of a destructured argument"
                )

            virtual_arg = replace(
                virtual_arg,
                destructure=self,
                has_value=False,
            )

            result.append(virtual_arg)

        return result


from cappa.command import Command, FinalCommand  # noqa: E402



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Any

import pytest
from typing_extensions import Annotated

import cappa
from cappa.default import DefaultFormatter
from tests.utils import Backend, CapsysOutput, backends, parse


@backends
def test_show_default_default_true(backend: Backend, capsys: Any):
    @dataclass
    class Command:
        foo: bool = False

    assert parse(Command, backend=backend).foo is False

    with pytest.raises(cappa.HelpExit):
        parse(Command, "--help", backend=backend)

    output = CapsysOutput.from_capsys(capsys)
    assert "(Default: False)" in output.stdout


@backends
def test_show_default_set_false(backend: Backend, capsys: Any):
    @dataclass
    class Command:
        foo: Annotated[bool, cappa.Arg(show_default=False)] = False

    assert parse(Command, backend=backend).foo is False

    with pytest.raises(cappa.HelpExit):
        parse(Command, "--help", backend=backend)

    output = CapsysOutput.from_capsys(capsys)
    assert "(Default: False)" not in output.stdout


@backends
def test_show_default_no_option_shows_for_false_default(backend: Backend, capsys: Any):
    @dataclass
    class Command:
        foo: Annotated[bool, cappa.Arg(long="--foo/--no-foo")] = False

    assert parse(Command, backend=backend).foo is False

    with pytest.raises(cappa.HelpExit):
        parse(Command, "--help", backend=backend)

    stdout = CapsysOutput.from_capsys(capsys).stdout
    # Bool flags are combined into single brackets
    assert "[--foo, --no-foo]" in stdout
    # Should show default False but not True (flexible whitespace between columns)
    assert re.search(r"\[--foo, --no-foo\]\s+\(Default:\s*False\)", stdout)
    assert "(Default: True)" not in stdout


@backends
def test_show_default_no_option_shows_for_true_default(backend: Backend, capsys: Any):
    @dataclass
    class Command:
        foo: Annotated[bool, cappa.Arg(long="--foo/--no-foo")] = True

    assert parse(Command, backend=backend).foo is True

    with pytest.raises(cappa.HelpExit):
        parse(Command, "--help", backend=backend)

    stdout = CapsysOutput.from_capsys(capsys).stdout
    # Bool flags with no value are combined into single brackets
    assert "[--foo, --no-foo]" in stdout
    # Should show default True but not False (flexible whitespace between columns)
    assert re.search(r"\[--foo, --no-foo\]\s+\(Default:\s*True\)", stdout)
    assert "(Default: False)" not in stdout


@backends
def test_explicit_arg_default(backend: Backend, capsys: Any):
    @dataclass
    class Command:
        foo: Annotated[bool, cappa.Arg(long="--foo/--no-foo", default=False)]

    with pytest.raises(cappa.HelpExit):
        parse(Command, "--help", backend=backend)

    stdout = CapsysOutput.from_capsys(capsys).stdout
    # Bool flags are combined into single brackets
    assert "[--foo, --no-foo]" in stdout
    # Flexible whitespace between columns
    assert re.search(r"\[--foo, --no-foo\]\s+\(Default:\s*False\)", stdout)


@backends
def test_show_default_string(backend: Backend, capsys: Any):
    @dataclass
    class Command:
        foo: Annotated[str, cappa.Arg(show_default="~{default}~")] = "asdf"

    with pytest.raises(cappa.HelpExit):
        parse(Command, "--help", backend=backend)

    stdout = CapsysOutput.from_capsys(capsys).stdout.replace(" ", "")
    assert "[FOO](Default:~asdf~)" in stdout


@backends
def test_show_default_explicit(backend: Backend, capsys: Any):
    @dataclass
    class Command:
        foo: Annotated[str, cappa.Arg(show_default=DefaultFormatter("!{default}!"))] = (
            "asdf"
        )

    with pytest.raises(cappa.HelpExit):
        parse(Command, "--help", backend=backend)

    stdout = CapsysOutput.from_capsys(capsys).stdout.replace(" ", "")
    assert "[FOO](Default:!asdf!)" in stdout


@backends
def test_static(backend: Backend, capsys: Any):
    @dataclass
    class Command:
        foo: Annotated[str | None, cappa.Arg(show_default="always show")] = None

    with pytest.raises(cappa.HelpExit):
        parse(Command, "--help", backend=backend)

    stdout = CapsysOutput.from_capsys(capsys).stdout
    assert re.search(r"\[FOO\]\s+\(Default: always show\)", stdout)



from __future__ import annotations

import contextlib
import importlib
import inspect
from typing import (
    Any,
    AsyncGenerator,
    Callable,
    Generator,
    Hashable,
    List,
    Sequence,
    cast,
)

from typing_extensions import TypeGuard

from cappa.class_inspect import has_command
from cappa.command import Command
from cappa.invoke.types import (
    C,
    Dep,
    DepTypes,
    InvokeCallable,
    InvokeCallableSpec,
    InvokeResolutionError,
    Resolved,
    SelfType,
)
from cappa.output import Output
from cappa.state import State
from cappa.type_view import CallableView, TypeView
from cappa.typing import find_annotations, get_method_class


def resolve_callable(
    command: Command[Any],
    parsed_command: Command[C],
    instance: C,
    *,
    implicit_deps: dict[Hashable, Any],
    output: Output,
    state: State[Any],
    deps: DepTypes = None,
) -> tuple[Resolved[C], Sequence[Resolved[Any]]]:
    try:
        fn: Callable[..., Any] = resolve_invoke_handler(parsed_command, implicit_deps)

        implicit_deps[Output] = output
        implicit_deps[Command] = parsed_command
        implicit_deps[State] = state
        implicit_deps[SelfType] = implicit_deps[cast(Hashable, parsed_command.cmd_cls)]

        global_deps = resolve_global_deps(deps, implicit_deps)

        fulfilled_deps: dict[Hashable, Any] = {**implicit_deps, **global_deps}
        resolved = fulfill_deps(fn, fulfilled_deps)
    except InvokeResolutionError as e:
        raise InvokeResolutionError(
            f"Failed to invoke {parsed_command.cmd_cls} due to resolution failure."
        ) from e

    return (resolved, tuple(global_deps.values()))


def resolve_global_deps(
    deps: DepTypes, implicit_deps: dict[Hashable, Any]
) -> dict[Hashable, Any]:
    result: dict[Hashable, Any] = {}

    if not deps:
        return result

    # Coerce the sequence variant of input into the mapping equivalent.
    if isinstance(deps, Sequence):
        deps = {cast(InvokeCallable[Any], d): Dep(d) for d in deps}

    for source_function, dep in deps.items():
        # Deps need to be fulfilled, whereas raw values are taken directly.
        if isinstance(dep, Dep):
            dep_callable = resolve_callable_reference(dep.callable)  # pyright: ignore
            value = fulfill_deps(dep_callable, implicit_deps)
        else:
            value = Resolved(source_function, result=dep)

        # We map back to the source dep, i.e. the key in the input mapping. This
        # translates to the raw function in the sequence input, or the source
        # dep being overridden in the dict input.
        source_dep: Dep[Any] = Dep(source_function)
        result[source_dep] = value

    return result


def resolve_invoke_handler(
    command: Command[C], implicit_deps: dict[Hashable, Any]
) -> Callable[..., C]:
    fn = command.invoke

    if not fn:
        command_type = cast(Hashable, command.cmd_cls)
        instance = implicit_deps.get(command_type)
        if callable(instance):
            return instance.__call__  # pyright: ignore

        raise InvokeResolutionError(
            f"Cannot call `invoke` for a command which does not have an invoke handler: {command.cmd_cls}."
        )

    return resolve_callable_reference(fn)


def resolve_callable_reference(fn: InvokeCallableSpec[C] | None) -> InvokeCallable[C]:
    if isinstance(fn, str):
        try:
            module_name, fn_name = fn.rsplit(".", 1)
        except ValueError:
            raise InvokeResolutionError(
                f"Invoke `{fn}` must be a fully qualified reference to a function in a module."
            )

        # Let this bubble upwards if the modules doesn't exist.
        try:
            module = importlib.import_module(module_name)
        except ModuleNotFoundError as e:
            name = getattr(e, "name", None) or str(e)
            raise InvokeResolutionError(
                f"No module '{name}' when attempting to load '{fn}'."
            )

        if not hasattr(module, fn_name):
            raise InvokeResolutionError(
                f"Module {module} does not have a function `{fn_name}`."
            )

        fn = getattr(module, fn_name)

    if not callable(fn):
        raise InvokeResolutionError(f"`{fn}` does not reference a valid callable.")

    return cast(Callable[..., Any], fn)


def fulfill_deps(
    fn: Callable[..., C], fulfilled_deps: dict[Hashable, Any], allow_empty: bool = False
) -> Resolved[C]:
    args: list[Any] = []
    result: dict[str, Any] = {}

    try:
        callable_view = CallableView.from_callable(fn, include_extras=True)
    except NameError as e:  # pragma: no cover
        name = getattr(e, "name", None) or str(e)
        raise InvokeResolutionError(
            f"Could not collect resolve reference to {name} for `{getattr(fn, '__name__', '')}`"
        )
    except (ValueError, AttributeError):
        # ValueError is common amongst builtins. Perhaps TypeView ought to be handling this.
        # AttributeError is currently an issue with Enums, I think TypeView should **definitely**
        # handle this.
        return Resolved(fn, result)

    for index, param_view in enumerate(callable_view.parameters):
        type_view: TypeView[Any] = param_view.type_view  # pyright: ignore

        # Unwrap TypeAliasType instances like `type Foo = Annotated[int, Dep(foo)]`
        type_view = type_view.strip_type_alias()

        # "Native" supported annotations
        if type_view.is_annotated and SelfType in type_view.metadata:
            result[param_view.name] = fulfilled_deps[SelfType]

        elif type_view.fallback_origin in fulfilled_deps:
            result[param_view.name] = fulfilled_deps[type_view.fallback_origin]

        # "Dep(foo)" annotations
        elif deps := cast(List[Dep[Any]], find_annotations(type_view, Dep)):
            assert len(deps) == 1
            dep = deps[0]

            # Whereas everything else should be a resolvable explicit Dep, which might have either
            # already been fullfullfilled, or yet need to be.
            if dep not in fulfilled_deps:
                fulfilled_deps[dep] = fulfill_deps(
                    cast(Callable[..., Any], dep.callable), fulfilled_deps
                )

            result[param_view.name] = fulfilled_deps[dep]

        # Method `self` arguments can be assumed to be typed as the literal class they reside inside,
        # These classes should always be already fulfilled by the root command structure.
        elif index == 0 and inspect.ismethod(fn):
            cls = get_method_class(fn)

            if has_command(fn.__self__):
                value = fulfilled_deps[cls]
                args.append(value)

        # If there's a default, we can just skip it and let the default fulfill the value.
        # Alternatively, `allow_empty` might be True to indicate we shouldn't error.
        elif param_view.has_default or allow_empty:
            continue

        # Non-annotated args are either implicit dependencies (and thus already fulfilled),
        # or arguments that we cannot fulfill and should error.
        else:
            param_annotation = (
                param_view.type_view.repr_type
                if param_view.has_annotation
                else "<empty>"
            )

            raise InvokeResolutionError(
                f"`{param_view.name}: {param_annotation}` "
                f"is not a valid dependency for Dep({fn.__name__})."
            )

    return Resolved(fn, kwargs=result, args=tuple(args))


def is_implicit_context_manager(
    value: Any,
) -> TypeGuard[Callable[..., Generator[Any, Any, Any]]]:
    return inspect.isgeneratorfunction(value)


def is_context_manager(value: Any) -> TypeGuard[contextlib.AbstractContextManager[Any]]:
    return isinstance(value, contextlib.AbstractContextManager)


def is_implicit_async_context_manager(
    value: Any,
) -> TypeGuard[Callable[..., AsyncGenerator[Any, Any]]]:
    return inspect.isasyncgenfunction(value)


def is_async_context_manager(
    value: Any,
) -> TypeGuard[contextlib.AbstractAsyncContextManager[Any]]:
    return isinstance(value, contextlib.AbstractAsyncContextManager)



from __future__ import annotations

from dataclasses import dataclass

import pytest
from typing_extensions import Annotated

import cappa
from cappa.output import Exit
from tests.utils import Backend, backends, parse


@backends
def test_valid(backend: Backend):
    @cappa.command(default_short=True)
    @dataclass
    class Command:
        foo: int = 4
        bar: int = 5

    test = parse(Command, backend=backend)
    assert test == Command(4, 5)

    test = parse(Command, "-f", "1", "-b", "2", backend=backend)
    assert test == Command(1, 2)


@backends
def test_override(backend: Backend):
    @cappa.command(default_short=True)
    @dataclass
    class Command:
        foo: int
        far: Annotated[int, cappa.Arg(short="-k")]

    test = parse(Command, "-f", "1", "-k", "2", backend=backend)
    assert test == Command(1, 2)


@backends
def test_invalid(backend: Backend):
    @cappa.command(default_short=True)
    @dataclass
    class Command:
        foo: int
        far: int

    with pytest.raises(Exit) as e:
        parse(Command, backend=backend)
    assert "conflicting option string: -f" in str(e.value.message).lower()



from __future__ import annotations

from dataclasses import dataclass
from typing import Union

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, invoke


def foo(tlc: TopLevelCommand, subcommand: Subcommand, foo: Foo):
    return (tlc, subcommand, foo)


@cappa.command(invoke=foo)
@dataclass
class Foo: ...


@dataclass
class Subcommand:
    cmd: Annotated[Foo, cappa.Subcommand]


@cappa.command(invoke=lambda: 4)
@dataclass
class TopLevelCommand:
    cmd: Annotated[Union[Subcommand, None], cappa.Subcommand] = None


@backends
def test_every_dependency_level(backend: Backend):
    result = invoke(TopLevelCommand, "subcommand", "foo", backend=backend)
    assert result == (
        TopLevelCommand(cmd=Subcommand(cmd=Foo())),
        Subcommand(cmd=Foo()),
        Foo(),
    )


@backends
def test_no_subcmd(backend: Backend):
    result = invoke(TopLevelCommand, backend=backend)
    assert result == 4



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    ClassVar,
    Hashable,
    Protocol,
    TextIO,
    Union,
    runtime_checkable,
)

import rich.prompt
from typing_extensions import Self, TypeAlias, TypeVar

from cappa.state import State
from cappa.type_view import Empty, EmptyType

T = TypeVar("T")


@dataclass(frozen=True)
class Default:
    """Represents the sequence of fallback methods of default value retrieval for an `Arg`.

    **All** argument defaults (including static values) are coerced into `Default` instance.

    When evaluated for an argument, the provided sequence of default retrieval methods
    are evaluated in order, returning the value of the first option to return a non-None value.
    The `Default.default` value is returned if none of the provided methods return a value.

    Examples:
    - `Arg(default=4) == Arg(default=Default(default=4))`
    - `Arg(default=Env("FOO")) == Arg(default=Default(Env("FOO")))`
    - `Arg(default=Env("FOO") | Env("BAR")) == Arg(default=Default(Env("FOO"), Env("BAR")))`
    """

    sequence: tuple[DefaultType, ...] = ()
    default: EmptyType | Any = Empty

    def __init__(self, *sequence: DefaultTypes, default: Any = Empty):
        object.__setattr__(self, "sequence", sequence)
        object.__setattr__(self, "default", default)

    @classmethod
    def from_value(cls, default: Any) -> Default:
        """Produce a `Default` instance where the provided value is an explicit default.

        A default produced this way will resolve to a `Value` which gets appended to the
        fallback sequence rather than being overwriting the Default-level static fallback.

        Essentially `Default(..., Value(4), default=5)` will **always** evaluate to `4` if
        upstream fallbacks do not provide a value, and `5` will never be used. In practice
        this is utilized when `Arg(default=4)` is provided.

        This is distinct but similar to `Default() | 4 | 5` (which uses `fallback_to`), which
        would resolve to `Default((), default=5)`, the 4 being overwritten and never used.
        """
        return cls().fallback_to(default, explicit=True)

    @classmethod
    def fallback_from(cls, *defaults: Default | EmptyType | Any | None) -> Default:
        """Return a `Default` instance from a sequence of potential defaults.

        This is used to compose different levels of default sources (Arg.default versus
        normalization time defaults).
        """
        for default in defaults:
            if default is not Empty:
                return Default.from_value(default)

        return cls()

    def fallback_to(self, other: DefaultTypes | Any, explicit: bool = False) -> Default:
        """Produce a new Default from the current default and a supplied alternative.

        Examples:
            >>> Default(Env("FOO")).fallback_to(5)
            Default(sequence=(Env(env_vars=('FOO',), ...),), default=5)

            >>> Default(Env("FOO")).fallback_to(5, explicit=True)
            Default(sequence=(Env(env_vars=('FOO',), ...), Value(value=5)), default=<_EmptyEnum.EMPTY: 0>)
        """
        cls = type(self)
        if isinstance(other, cls):
            return cls(*self.sequence, *other.sequence, default=other.default)

        if isinstance(other, default_types):
            default = self.default
            if self.default is Empty and isinstance(other, Env):
                default = other.default

            if isinstance(other, PromptType):
                other = Prompt.from_prompt(other)

            if isinstance(other, ConfirmType):
                other = Confirm.from_confirm(other)

            return cls(*self.sequence, other, default=default)

        if explicit:
            return cls(*self.sequence, Value(other))

        return cls(*self.sequence, default=other)

    def __or__(self, other: Any) -> Default:
        """Compose two potential defaults, returning a new one.

        For example, `Default() | Prompt('...') | Env('FOO') | 5`. See `fallback_to` for details.
        """
        return self.fallback_to(other)

    def __call__(
        self, state: State[Any] | None = None, input: TextIO | None = None
    ) -> tuple[bool, Any]:
        """Evaluate the default retrieval sequence, returning the first non-Empty value."""
        for default in self.sequence:
            if isinstance(default, ValueFrom):
                value = default(state=state)
            elif isinstance(default, (Prompt, Confirm)):
                value = default(input=input)
            else:
                value = default()

            if value is not Empty:
                return default.is_parsed, value

        if self.default is Empty:
            return True, None

        return True, self.default

    @property
    def has_value(self) -> bool:
        """Whether the default instance **has** a default or if it's Empty."""
        return not self.sequence and self.default is Empty

    @property
    def fallback_value(self) -> Any:
        return next(
            (d.value for d in self.sequence if isinstance(d, Value)), self.default
        )


@runtime_checkable
class DefaultType(Protocol):
    is_parsed: ClassVar[bool] = False

    def __call__(self, *_: Any, **__: Any) -> Any: ...  # pragma: no cover

    def __or__(self, other: DefaultTypes) -> Default:
        return Default(self) | other


@dataclass(frozen=True)
class Env(DefaultType):
    """Lazily evaluates environment variables in the order given.

    Argument value interpolation will handle and `Arg`'s default being an `Env`
    instance, by evaluating the `Env`, in the event the parser-level value
    fell back to the default.

    Examples:
        >>> from cappa import Arg, Env
        >>> arg = Arg(default=Env("HOST", "HOSTNAME", default="localhost"))
    """

    env_vars: tuple[str, ...]
    default: str | EmptyType | None = Empty

    def __init__(
        self, env_var: str, *env_vars: str, default: str | EmptyType | None = Empty
    ):
        object.__setattr__(self, "env_vars", (env_var, *env_vars))
        object.__setattr__(self, "default", default)

    def __call__(self) -> str | EmptyType | None:
        for env_var in self.env_vars:
            value = os.getenv(env_var)
            if value is not None:
                return value

        if self.default is None:
            return Empty
        return self.default


class Prompt(rich.prompt.Prompt, DefaultType):
    """Prompt the user for a value, returning the response.

    Examples:
        >>> from cappa import Arg, Prompt
        >>> arg = Arg(default=Prompt("Ask user for value"))
    """

    @classmethod
    def from_prompt(cls, prompt: Prompt | PromptType) -> Self:
        if isinstance(prompt, cls):
            return prompt
        return cls(
            console=prompt.console,
            prompt=prompt.prompt,
            password=prompt.password,
            choices=prompt.choices,
            show_default=prompt.show_default,
            show_choices=prompt.show_choices,
        )

    def __call__(self, input: TextIO | None = None):  # type: ignore
        return super().__call__(default=Empty, stream=input)


class Confirm(rich.prompt.Confirm, DefaultType):
    """Prompt the user for a confirmation, returning `True`/`False`.

    Examples:
        >>> from cappa import Arg, Confirm
        >>> arg = Arg(default=Confirm("Confirm with user"))
    """

    @classmethod
    def from_confirm(cls, confirm: Confirm | ConfirmType) -> Self:
        if isinstance(confirm, cls):
            return confirm
        return cls(
            console=confirm.console,
            prompt=confirm.prompt,
            password=confirm.password,
            choices=confirm.choices,
            show_default=confirm.show_default,
            show_choices=confirm.show_choices,
        )

    def __call__(self, input: TextIO | None = None):  # type: ignore
        return super().__call__(default=Empty, stream=input)


@dataclass(frozen=True)
class Value(DefaultType):
    value: Any

    is_parsed: ClassVar[bool] = True

    def __call__(self, state: State[Any] | None = None):
        return self.value


@dataclass(frozen=True)
class ValueFrom(DefaultType):
    """Retrieve the default value by calling the provided function.

    Optionally, `*args` and `**kwargs` to the supplied function can be
    provided ahead of time, to be supplied in the event the function is called.

    Examples:
        >>> from cappa import Arg, ValueFrom
        >>> def from_file(name):
        ...     with open(name) as f:
        ...         return f.read()
        >>>
        >>> arg = Arg(default=ValueFrom(from_file, name="config.json"))
    """

    callable: Callable[..., Any]
    kwargs: dict[str, Any]

    is_parsed: ClassVar[bool] = True

    def __init__(self, callable: Callable[..., Any], **kwargs: Any):
        object.__setattr__(self, "callable", callable)
        object.__setattr__(self, "kwargs", kwargs)

    def __call__(self, state: State[Any] | None = None):
        from cappa.invoke.base import fulfill_deps

        kwargs = self.kwargs
        if state:
            deps: dict[Hashable, Any] = {State: state}
            resolved = fulfill_deps(self.callable, deps, allow_empty=True)
            kwargs = {**resolved.kwargs, **self.kwargs}

        return self.callable(**kwargs)


@dataclass
class DefaultFormatter:
    format: str = "{default}"
    show: bool | None = True

    @classmethod
    def disabled(cls) -> Self:
        return cls(show=False)

    @classmethod
    def from_unknown(cls, value: str | bool | Self) -> Self:
        if isinstance(value, cls):
            return value

        if isinstance(value, str):
            return cls(format=value)

        return cls(show=bool(value))

    def format_default(self, default: Default, default_format: str = "") -> str:
        if self.show is False:
            return ""

        default_raw_value = default.fallback_value
        if default_raw_value in (None, Empty):
            default_raw_value = ""

        default_formatted_value = self.format.format(default=default_raw_value)

        if default_formatted_value == "":
            return default_formatted_value

        return default_format.format(default=default_formatted_value)


PromptType = rich.prompt.Prompt
ConfirmType = rich.prompt.Confirm
default_types = (rich.prompt.Prompt, rich.prompt.Confirm, Env, ValueFrom)
DefaultTypes: TypeAlias = Union[
    Default, DefaultType, rich.prompt.Prompt, rich.prompt.Confirm
]



from __future__ import annotations

from dataclasses import dataclass
from typing import Union

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_(backend: Backend):
    @dataclass
    class Args:
        foo: Annotated[bool, cappa.Arg(long=True)] = False
        raw: Annotated[Union[list[str], None], cappa.Arg(num_args=-1)] = None

    test = parse(Args, "--foo", "--", "value", backend=backend)
    assert test.foo is True
    assert test.raw == ["value"]

    test = parse(Args, "--", "--foo", "value", backend=backend)
    assert test.foo is False
    assert test.raw == ["--foo", "value"]



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import parse


@dataclass
class Other:
    arg2: Annotated[int, cappa.Arg(long=True)] = 4
    arg3: Annotated[str, cappa.Arg(long=True)] = "value3"


@dataclass
class CLI:
    other: cappa.Destructured[Other]


def test_destructured_implicit_default():
    result = parse(CLI)
    assert result == CLI(other=Other())


@dataclass
class OptionalCLI:
    other: cappa.Destructured[Other | None]


def test_destructured_implicit_default_optional_annotation():
    result = parse(OptionalCLI)
    assert result == OptionalCLI(other=None)



from __future__ import annotations

from dataclasses import dataclass
from unittest.mock import patch

import pytest
from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_missing_default(backend: Backend):
    @dataclass
    class ArgTest:
        default: bool

    test = parse(ArgTest, backend=backend)
    assert test.default is False

    test = parse(ArgTest, "--default", backend=backend)
    assert test.default is True


@backends
def test_default(backend: Backend):
    @dataclass
    class ArgTest:
        default: bool = False

    test = parse(ArgTest, backend=backend)
    assert test.default is False

    test = parse(ArgTest, "--default", backend=backend)
    assert test.default is True


@backends
def test_explicit_short(backend: Backend):
    @dataclass
    class ArgTest:
        explicit_short: Annotated[bool, cappa.Arg(short="-e")] = False

    test = parse(ArgTest, backend=backend)
    assert test.explicit_short is False

    test = parse(ArgTest, "-e", backend=backend)
    assert test.explicit_short is True


@backends
def test_explicit_long(backend: Backend):
    @dataclass
    class ArgTest:
        explicit_long: Annotated[bool, cappa.Arg(long="--meow")] = False

    test = parse(ArgTest, backend=backend)
    assert test.explicit_long is False

    test = parse(ArgTest, "--meow", backend=backend)
    assert test.explicit_long is True


@backends
def test_store_false(backend: Backend):
    @dataclass
    class ArgTest:
        default_true: Annotated[bool, cappa.Arg(long="--false")] = True

    test = parse(ArgTest, backend=backend)
    assert test.default_true is True

    test = parse(ArgTest, "--false", backend=backend)
    assert test.default_true is False


@backends
def test_true_false_option(backend: Backend):
    @dataclass
    class ArgTest:
        true_false: Annotated[bool, cappa.Arg(long="--true/--no-true")] = True

    test = parse(ArgTest, backend=backend)
    assert test.true_false is True

    test = parse(ArgTest, "--true", backend=backend)
    assert test.true_false is True

    test = parse(ArgTest, "--no-true", backend=backend)
    assert test.true_false is False

    with pytest.raises(cappa.Exit) as e:
        parse(ArgTest, "--garbage", backend=backend)

    assert e.value.code == 2
    assert "unrecognized arguments: --garbage" in str(e.value.message).lower()


@backends
def test_optional_bool(backend: Backend):
    """Optional bool with default should still infer store_true/false action."""

    @dataclass
    class ArgTest:
        true_false: Annotated[bool | None, cappa.Arg(long="--true/--no-true")] = None

    test = parse(ArgTest, backend=backend)
    assert test.true_false is None

    test = parse(ArgTest, "--true", backend=backend)
    assert test.true_false is True

    test = parse(ArgTest, "--no-true", backend=backend)
    assert test.true_false is False


@backends
def test_optional_bool_no_default(backend: Backend):
    """Optional bool should default to None rather than bool."""

    @dataclass
    class ArgTest:
        true_false: Annotated[bool | None, cappa.Arg(long="--true/--no-true")]

    test = parse(ArgTest, backend=backend)
    assert test.true_false is None


@backends
def test_optional_bool_and_int(backend: Backend):
    """Union of multiple types, including bool does not infer bool action."""

    @dataclass
    class ArgTest:
        true_false: Annotated[bool | int | None, cappa.Arg(short=True)]

    test = parse(ArgTest, backend=backend)
    assert test.true_false is None

    test = parse(ArgTest, "-t", "4", backend=backend)
    assert test.true_false == 4

    test = parse(ArgTest, "-t", "asdf", backend=backend)
    assert test.true_false is True


@backends
def test_env_default_value_precedence(backend: Backend):
    """Assert a bool flag yields correct value given an env default."""

    @dataclass
    class ArgTest:
        env_default: Annotated[
            bool, cappa.Arg(long=True, default=cappa.Env("ENV_DEFAULT"))
        ] = False

    test = parse(ArgTest, backend=backend)
    assert test.env_default is False

    with patch("os.environ", new={"ENV_DEFAULT": "1"}):
        test = parse(ArgTest, backend=backend)
    assert test.env_default is True

    test = parse(ArgTest, "--env-default", backend=backend)
    assert test.env_default is True

    with patch("os.environ", new={"ENV_DEFAULT": "1"}):
        test = parse(ArgTest, "--env-default", backend=backend)
    assert test.env_default is True


@backends
def test_sole_no_arg(backend: Backend):
    @dataclass
    class ArgTest:
        no_dry_run: bool = False

    test = parse(ArgTest, backend=backend)
    assert test.no_dry_run is False

    test = parse(ArgTest, "--no-dry-run", backend=backend)
    assert test.no_dry_run is True


@backends
def test_sole_no_arg_inverted(backend: Backend):
    @dataclass
    class ArgTest:
        no_dry_run: bool = True

    test = parse(ArgTest, backend=backend)
    assert test.no_dry_run is True

    test = parse(ArgTest, "--no-dry-run", backend=backend)
    assert test.no_dry_run is False



from __future__ import annotations

import asyncio
from dataclasses import dataclass
from typing import Any

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, invoke_async


def bar():
    return 7


async def foo(bar: Annotated[int, cappa.Dep(bar)]):
    return bar + 9


async def handler(foo: Annotated[int, cappa.Dep(foo)]):
    return foo + 1


@cappa.command(invoke=handler)
@dataclass
class Command: ...


def idk():
    print("wat")


@backends
def test_async_fn(backend: Backend, capsys: Any):
    result = asyncio.run(invoke_async(Command, backend=backend, deps=[idk]))
    assert result == 7 + 9 + 1

    assert capsys.readouterr().out == "wat\n"



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Union

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_list_option(backend: Backend):
    @dataclass
    class ArgTest:
        variable_number: Annotated[list[str], cappa.Arg(short=True, long=True)] = field(
            default_factory=lambda: []
        )

    test = parse(
        ArgTest,
        "-v",
        "one",
        "--variable-number",
        "two",
        backend=backend,
    )
    assert test.variable_number == ["one", "two"]


@backends
def test_list_positional(backend: Backend):
    @dataclass
    class ArgTest:
        variable_number: list[str] = field(default_factory=lambda: [])

    test = parse(
        ArgTest,
        "one",
        "two",
        "three",
        backend=backend,
    )
    assert test.variable_number == ["one", "two", "three"]


@backends
def test_optional_list(backend: Backend):
    @dataclass
    class ArgTest:
        value: Annotated[Union[list[str], None], cappa.Arg(short=True, long=True)] = (
            None
        )

    test = parse(
        ArgTest,
        "--value=one",
        "--value=two",
        backend=backend,
    )
    assert test.value == ["one", "two"]

    test = parse(ArgTest, backend=backend)
    assert test.value is None



from __future__ import annotations

from dataclasses import dataclass
from unittest.mock import patch

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@dataclass
class Command:
    opt: Annotated[str | None, cappa.Arg(default=cappa.Env("DOESNT_EXIST"))]
    opt2: Annotated[
        str | None, cappa.Arg(default=cappa.Env("DOESNT_EXIST", default=None))
    ]
    opt3: Annotated[
        str | None, cappa.Arg(default=cappa.Env("DOESNT_EXIST", default=None))
    ] = None


@backends
def test_env_missing(backend: Backend):
    test = parse(Command, backend=backend)
    assert test == Command(None, None, None)


@backends
def test_default_is_not_mapped(backend: Backend):
    with patch("os.environ", new={"DOESNT_EXIST": "asdf"}):
        test = parse(Command, backend=backend)

    assert test == Command("asdf", "asdf", "asdf")



from __future__ import annotations

import dataclasses
from typing import TYPE_CHECKING, Any, Iterable, Mapping, TextIO

from typing_extensions import Annotated, TypeAlias

from cappa.arg import Group
from cappa.class_inspect import Field, extract_dataclass_metadata
from cappa.completion.types import Completion
from cappa.invoke.types import Resolved
from cappa.state import State
from cappa.type_view import Empty, EmptyType, TypeView
from cappa.typing import T, assert_type, find_annotations

if TYPE_CHECKING:
    from cappa.arg import FinalArg
    from cappa.command import Alias, Command, FinalCommand
    from cappa.help import HelpFormattable
    from cappa.output import Output


DEFAULT_SUBCOMMAND_GROUP = Group(3, "Subcommands", section=1)


@dataclasses.dataclass
class Subcommand:
    """Describe a CLI subcommand.

    Arguments:
        field_name: Defaults to the name of the class, converted to dash case, but
            can be overridden here.
        required: Defaults to automatically inferring requiredness, based on whether the
            class's value has a default. By setting this, you can force a particular value.
        group: The subcommand group, for use in controlling help text for the subcommand, and
            where it is displayed. This can be any of: the string name (``'Subcommands'``),
            a 2-tuple of the `order` and the name (``(3, "Subcommands")``), or a :class:`Group`
            instance (``Group(3, 'Subcommands')``)
        hidden: Whether the argument should be hidden in help text. Defaults to False.
        options: A mapping of the subcommand names to the corresponding `Command` to which
            the subcommands refer. Unless imperatively constructing the CLI structure, this
            field should generally always be inferred automatically.
        types: Defaults to the class's annotated types, but can be overridden here.
    """

    field_name: str | EmptyType = Empty
    required: bool | None = None
    group: str | tuple[int, str] | Group = DEFAULT_SUBCOMMAND_GROUP
    hidden: bool = False

    options: Mapping[str, Command[Any]] = dataclasses.field(default_factory=lambda: {})
    types: Iterable[type] | EmptyType = Empty

    # Mapping of alias name -> (canonical name in `options`, Alias metadata).
    # Populated during `normalize` from each Command's `aliases` field; collisions raise.
    alias_map: dict[str, tuple[str, Alias]] = dataclasses.field(
        default_factory=lambda: {}
    )

    @classmethod
    def detect(cls, field: Field, type_view: TypeView[Any]) -> Subcommand | None:
        subcommands = find_annotations(type_view, Subcommand) or None

        field_metadata = extract_dataclass_metadata(field, Subcommand)
        if field_metadata:
            subcommands = field_metadata

        if not subcommands:
            return None

        assert len(subcommands) == 1
        return subcommands[0]

    def normalize(
        self,
        type_view: TypeView[Any] | None = None,
        field_name: str | None = None,
        help_formatter: HelpFormattable | None = None,
        propagated_arguments: list[FinalArg[Any]] | None = None,
        state: State[Any] | None = None,
    ) -> FinalSubcommand:
        if type_view is None:
            type_view = TypeView(Any)

        field_name = field_name or assert_type(self.field_name, str)
        types = infer_types(self, type_view)
        required = infer_required(self, type_view)
        options = infer_options(
            self,
            types,
            help_formatter=help_formatter,
            propagated_arguments=propagated_arguments,
            state=state,
        )
        alias_map = build_alias_map(options)
        group = infer_group(self)

        return FinalSubcommand(
            hidden=self.hidden,
            field_name=field_name,
            types=types,
            required=required,
            options=options,
            alias_map=alias_map,
            group=group,
        )


@dataclasses.dataclass
class FinalSubcommand(Subcommand):
    """Post-normalization form of :class:`Subcommand` with narrowed field types.

    Produced exclusively by :meth:`Subcommand.normalize`.
    """

    field_name: str = ""  # pyright: ignore
    required: bool = False  # pyright: ignore
    group: Group = dataclasses.field(default_factory=lambda: DEFAULT_SUBCOMMAND_GROUP)  # pyright: ignore
    types: Iterable[type] = dataclasses.field(default_factory=tuple)  # pyright: ignore
    options: Mapping[str, FinalCommand[Any]] = dataclasses.field(  # pyright: ignore
        default_factory=dict
    )

    def resolve_name(self, name: str) -> str | None:
        """Return the canonical name for a typed-in name (canonical or alias).

        Returns `None` when `name` matches neither.
        """
        if name in self.options:
            return name
        if name in self.alias_map:
            return self.alias_map[name][0]
        return None

    def map_result(
        self,
        prog: str,
        parsed_args: dict[str, Any],
        *,
        output: Output,
        state: State[Any] | None = None,
        input: TextIO | None = None,
    ) -> tuple[Resolved[Any], dict[Any, Any]]:
        canonical = parsed_args.pop("__name__")
        option = self.options[canonical]
        return option.map_result(
            option, prog, parsed_args, output=output, state=state, input=input
        )

    def available_options(self) -> list[FinalCommand[Any]]:
        return [o for o in self.options.values() if not o.hidden]

    def names(self) -> list[str]:
        return [n for n, o in self.options.items() if not o.hidden]

    def visible_aliases_for(self, canonical: str) -> list[Alias]:
        """Visible (non-hidden) aliases for the given canonical subcommand name."""
        command = self.options.get(canonical)
        if command is None:
            return []
        return [a for a in command.resolved_aliases() if not a.hidden]

    def all_visible_names(self) -> list[str]:
        """Canonical names plus visible aliases, in declaration order."""
        result: list[str] = []
        for name in self.names():
            result.append(name)
            for alias in self.visible_aliases_for(name):
                result.append(alias.name)
        return result

    def names_str(self, delimiter: str = ", ") -> str:
        return delimiter.join(self.names())

    def completion(self, partial: str):
        return [Completion(o) for o in self.all_visible_names() if partial in o]


def infer_types(arg: Subcommand, type_view: TypeView[Any]) -> Iterable[type]:
    if arg.types is not Empty:
        return arg.types

    if type_view.is_union:
        return tuple(t.annotation for t in type_view.inner_types if not t.is_none_type)

    return (type_view.annotation,)


def infer_required(arg: Subcommand, annotation: TypeView[Any]) -> bool:
    if arg.required is not None:
        return arg.required

    return not annotation.is_optional


def infer_options(
    arg: Subcommand,
    types: Iterable[type],
    help_formatter: HelpFormattable | None = None,
    propagated_arguments: list[FinalArg[Any]] | None = None,
    state: State[Any] | None = None,
) -> dict[str, FinalCommand[Any]]:
    from cappa.command import Command

    if arg.options:
        return {
            name: type_command.collect(
                propagated_arguments=propagated_arguments,
                state=state,
            )
            for name, type_command in arg.options.items()
        }

    options: dict[str, FinalCommand[Any]] = {}
    for type_ in types:
        type_command: Command[Any] = Command.get(type_, help_formatter=help_formatter)  # pyright: ignore
        type_name = type_command.real_name()
        options[type_name] = type_command.collect(
            propagated_arguments=propagated_arguments
        )

    return options


def build_alias_map(
    options: dict[str, FinalCommand[Any]],
) -> dict[str, tuple[str, Alias]]:
    """Build alias name -> (canonical, Alias) for a set of subcommand options.

    Raises `ValueError` on collisions: an alias matching another command's
    canonical name, an alias colliding with another alias, or an alias that
    duplicates its own canonical name.
    """
    alias_map: dict[str, tuple[str, Alias]] = {}
    for canonical, command in options.items():
        for alias in command.resolved_aliases():
            if alias.name == canonical:
                raise ValueError(
                    f"Subcommand '{canonical}' has an alias matching its own name."
                )
            if alias.name in options:
                raise ValueError(
                    f"Subcommand alias '{alias.name}' (for '{canonical}') "
                    f"collides with another subcommand's name."
                )
            if alias.name in alias_map:
                other_canonical, _ = alias_map[alias.name]
                raise ValueError(
                    f"Subcommand alias '{alias.name}' is declared on both "
                    f"'{other_canonical}' and '{canonical}'."
                )
            alias_map[alias.name] = (canonical, alias)
    return alias_map


def infer_group(arg: Subcommand) -> Group:
    name = None
    order = 3

    if isinstance(arg.group, Group):
        return arg.group

    if isinstance(arg.group, str):
        name = arg.group

    if isinstance(arg.group, tuple):
        order, name = arg.group

    assert name
    return Group(name=name, order=order)


Subcommands: TypeAlias = Annotated[T, Subcommand]
DEFAULT_SUBCOMMAND = Subcommand()



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, List

import pytest

import cappa
from cappa.output import Exit
from cappa.parse import parse_list, parse_value
from tests.utils import Backend, backends, parse


@dataclass
class Foo:
    bar: str
    baz: list[int]


command = cappa.Command(
    Foo,
    arguments=[
        cappa.Arg(field_name="bar"),
        cappa.Arg(field_name="baz", parse=parse_list(List[int]), num_args=-1),
    ],
    help="Short help.",
    description="Long description.",
)


@backends
def test_valid(backend: Backend):
    result = parse(command, "one", "2", "3", backend=backend)
    assert result == Foo(bar="one", baz=[2, 3])


@pytest.mark.help
@backends
def test_help(capsys: Any, backend: Backend):
    with pytest.raises(Exit):
        parse(command, "-h", backend=backend)

    out = capsys.readouterr().out
    assert "Short help." in out
    assert "Long description." in out


@dataclass
class Bar:
    bar: str


@dataclass
class Foo2:
    sub: Bar


@backends
def test_subcommand(backend: Backend):
    command = cappa.Command(
        Foo2,
        arguments=[
            cappa.Subcommand(
                field_name="sub",
                options={
                    "bar": cappa.Command(
                        Bar,
                        arguments=[
                            cappa.Arg(field_name="bar", parse=parse_value(str)),
                        ],
                    )
                },
            ),
        ],
        help="Short help.",
        description="Long description.",
    )

    result = parse(command, "bar", "one", backend=backend)
    assert result == Foo2(sub=Bar("one"))



from cappa.base import collect, command, invoke, invoke_async, parse, parse_async
from cappa.command import Alias, Command, FinalCommand
from cappa.completion.types import Completion
from cappa.default import Confirm, Default, Env, Prompt, ValueFrom
from cappa.file_io import FileMode
from cappa.help import HelpFormattable, HelpFormatter
from cappa.invoke.types import Dep, Self
from cappa.output import Exit, HelpExit, Output
from cappa.parse import default_parse, unpack_arguments
from cappa.state import State
from cappa.subcommand import FinalSubcommand, Subcommand, Subcommands
from cappa.type_view import Empty, EmptyType

# isort: split
from cappa.arg import Arg, ArgAction, FinalArg, Group, NumArgs
from cappa.destructure import Destructured

# isort: split
from cappa import argparse
from cappa.parser import backend

__all__ = [
    "Alias",
    "Arg",
    "ArgAction",
    "Command",
    "Completion",
    "Confirm",
    "Default",
    "Dep",
    "Destructured",
    "Empty",
    "EmptyType",
    "Env",
    "Exit",
    "FileMode",
    "FinalArg",
    "FinalCommand",
    "FinalSubcommand",
    "Group",
    "HelpExit",
    "HelpFormattable",
    "HelpFormatter",
    "NumArgs",
    "Output",
    "Prompt",
    "Self",
    "State",
    "Subcommand",
    "Subcommands",
    "ValueFrom",
    "argparse",
    "backend",
    "collect",
    "command",
    "default_parse",
    "invoke",
    "invoke_async",
    "parse",
    "parse_async",
    "unpack_arguments",
]



from __future__ import annotations

import asyncio
from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends


def sync_parse_lowercase(value: str) -> str:
    return value.lower()


async def async_parse_uppercase(value: str) -> str:
    await asyncio.sleep(0.001)
    return value.upper()


@dataclass
class Command:
    sync_value: Annotated[str, cappa.Arg(parse=sync_parse_lowercase)]
    async_value: Annotated[str, cappa.Arg(parse=async_parse_uppercase)]

    def __call__(self):
        return self


@backends
def test(backend: Backend):
    result = asyncio.run(
        cappa.invoke_async(Command, argv=["HELLO", "world"], backend=backend)
    )
    assert result.sync_value == "hello"
    assert result.async_value == "WORLD"



from __future__ import annotations

import sys
from typing import Optional

import pytest
from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse

if sys.version_info < (3, 14):
    from pydantic import BaseModel, Field, dataclasses

    class PydanticCommand(BaseModel):
        name: str
        foo: Annotated[int, cappa.Arg(short=True)]

    @backends
    def test_base_model(backend: Backend):
        result = parse(PydanticCommand, "meow", "-f", "4", backend=backend)
        assert result == PydanticCommand(name="meow", foo=4)

    @dataclasses.dataclass
    class DataclassCommand:
        name: str
        foo: Annotated[int, cappa.Arg(short=True)]

    @backends
    def test_dataclass(backend: Backend):
        result = parse(DataclassCommand, "meow", "-f", "4", backend=backend)
        assert result == DataclassCommand(name="meow", foo=4)

    class OptSub(BaseModel):
        name: Optional[str] = None

    class OptionalSubcommand(BaseModel):
        sub: cappa.Subcommands[Optional[OptSub]] = None

    @backends
    def test_optional_subcommand(backend: Backend):
        result = parse(OptionalSubcommand, backend=backend)
        assert result == OptionalSubcommand(sub=None)

        result = parse(OptionalSubcommand, "opt-sub", backend=backend)
        assert result == OptionalSubcommand(sub=OptSub(name=None))

        result = parse(OptionalSubcommand, "opt-sub", "foo", backend=backend)
        assert result == OptionalSubcommand(sub=OptSub(name="foo"))

    @backends
    def test_class_validation(backend: Backend):
        class Foo(BaseModel):
            bar: Optional[int] = Field(gt=0)

        result = parse(Foo, "3", backend=backend)
        assert result.bar == 3

        with pytest.raises(cappa.Exit) as e:
            parse(Foo, "0", backend=backend)

        assert e.value.code == 2
        assert "greater than 0" in str(e.value.message)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, Generic, TypedDict, TypeVar, Union, overload


class BaseTypedDict(TypedDict): ...


S = TypeVar("S", bound=Union[Dict[str, Any], BaseTypedDict])


@dataclass
class State(Generic[S]):
    """A store of arbitrary data.

    State can be used as arguments to `Arg.parse`, `Arg.action`,
    or `Arg.default=ValueFrom(...)` callables, as a way to propagate shared
    state between different parts of the overall parse execution.

    Example:
        >>> from typing_extensions import Annotated
        >>> import cappa

        >>> def parse_arg(state: State):
        ...     state.state["arg"] = state.state.get("arg", 0)
        ...     return state.state["arg"]

        >>> @dataclass
        ... class Command:
        ...     arg1: Annotated[int, cappa.Arg(action=parse_arg)]
        ...     arg2: Annotated[int, cappa.Arg(parse=parse_arg)]
        ...     arg3: Annotated[int, cappa.Arg(default=cappa.ValueFrom(parse_arg))]
    """

    state: S = field(default_factory=lambda: {})  # type: ignore

    def set(self, key: str, value: Any):
        self.state[key] = value  # type: ignore

    def get(self, key: str, *, default: Any = None):
        return self.state.get(key, default)

    @overload
    @classmethod
    def ensure(cls, state: None) -> State[dict[str, Any]]: ...

    @overload
    @classmethod
    def ensure(cls, state: State[S]) -> State[S]: ...

    @classmethod
    def ensure(cls, state: State[S] | None) -> State[S] | State[dict[str, Any]]:
        if state is None:
            return State()
        return state



from __future__ import annotations

from dataclasses import dataclass
from typing import Literal

from typing_extensions import TypeAliasType

from tests.utils import Backend, backends, parse

Numbers = TypeAliasType("Numbers", Literal[1, 2, 3])


@backends
def test_manual_type_alias_type(backend: Backend):
    @dataclass
    class ArgTest:
        value: Numbers

    test = parse(ArgTest, "1", backend=backend)
    assert test.value == 1



from __future__ import annotations

import textwrap
from dataclasses import dataclass
from typing import Literal, Union

import pytest

import cappa
from tests.utils import Backend, backends, parse


@dataclass
class ArgTest:
    name: Union[int, Literal["one"]]


@backends
def test_valid_int(backend: Backend):
    test = parse(ArgTest, "5", backend=backend)
    assert test.name == 5


@backends
def test_valid_literal(backend: Backend):
    test = parse(ArgTest, "one", backend=backend)
    assert test.name == "one"


@backends
def test_invalid_string(backend: Backend):
    with pytest.raises(cappa.Exit) as e:
        parse(ArgTest, "thename", backend=backend)

    assert e.value.code == 2

    err = textwrap.dedent(
        """\
        Invalid value for 'name': Possible variants
         - Literal['one']: Invalid choice: 'thename' (choose from 'one')
         - int: invalid literal for int() with base 10: 'thename'"""
    )
    assert err in str(e.value.message)



from __future__ import annotations

import pathlib
from dataclasses import dataclass, field

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_positional_string(backend: Backend):
    @dataclass
    class ArgTest:
        name: str

    test = parse(ArgTest, "thename", backend=backend)
    assert test.name == "thename"


@backends
def test_path(backend: Backend):
    @dataclass
    class ArgTest:
        name: pathlib.PurePath

    test = parse(ArgTest, "./file.txt", backend=backend)
    assert test.name == pathlib.PurePath("./file.txt")


@backends
def test_bool(backend: Backend):
    @dataclass
    class ArgTest:
        flag: bool

    test = parse(ArgTest, "--flag", backend=backend)
    assert test.flag is True


@backends
def test_tuple(backend: Backend):
    @dataclass
    class ArgTest:
        exact_num_args: tuple[str, int] = field(
            default=("a", 0), metadata={"cappa": cappa.Arg(long=True)}
        )

    test = parse(
        ArgTest,
        "--exact-num-args",
        "three",
        "4",
        backend=backend,
    )
    assert test.exact_num_args == ("three", 4)


@backends
def test_missing_value_defaults(backend: Backend):
    @dataclass
    class ArgTest:
        exact_num_args: tuple[str, int] = field(
            default=("a", 0), metadata={"cappa": cappa.Arg(long=True)}
        )

    test = parse(ArgTest, backend=backend)
    assert test.exact_num_args == ("a", 0)



from __future__ import annotations

import contextlib
import inspect
from dataclasses import dataclass, field
from typing import (
    Any,
    AsyncGenerator,
    Callable,
    Coroutine,
    Generator,
    Generic,
    Mapping,
    Sequence,
    TypeVar,
    Union,
    cast,
)

from typing_extensions import Annotated

from cappa.output import Exit, Output
from cappa.type_view import Empty, EmptyType


class SelfType: ...


C = TypeVar("C")
T = TypeVar("T")
InvokeCallable = Callable[..., T]
InvokeCallableSpec = Union[InvokeCallable[T], str]


@dataclass(frozen=True)
class Dep(Generic[T]):
    """Describes the callable required to fulfill a given dependency."""

    callable: InvokeCallableSpec[T]


DepTypes = Union[
    Sequence[InvokeCallableSpec[Any]],
    Mapping[InvokeCallableSpec[Any], Union[Dep[Any], InvokeCallableSpec[Any], Any]],
    None,
]
Self = Annotated[T, SelfType]


class InvokeResolutionError(RuntimeError):
    """Raised for errors encountered during evaluation of invoke dependencies."""


@dataclass
class Resolved(Generic[C]):
    callable: InvokeCallableSpec[C]
    kwargs: dict[str, Any | Resolved[Any]] = field(default_factory=lambda: {})
    args: tuple[Any, ...] = field(default=())
    result: C | EmptyType = Empty

    def call(self, *args: Any, output: Output | None = None, managed: bool = True):
        with self.get(*args, output=output, managed=managed) as value:
            return value

    async def call_async(
        self, *args: Any, output: Output | None = None, managed: bool = True
    ):
        async with self.get_async(*args, output=output, managed=managed) as value:
            return value

    @contextlib.contextmanager
    def get(
        self, *args: Any, output: Output | None = None, managed: bool = True
    ) -> Generator[C, None, None]:
        """Get the resolved value.

        The value itself is cached in the event it's used as a dependency to more
        than one dependency.
        """
        if self.result is not Empty:
            yield self.result
            return

        with contextlib.ExitStack() as stack:
            # Non-resolved values are literal values that can be recorded directly.
            finalized_kwargs = dict(self.iter_kwargs(is_resolved=False))

            # Resolved values need to be recursed into. In order to handle the
            # wrapping context manager, we need to enter all contexts, and only
            # exit at the end.
            for k, v in self.iter_kwargs(is_resolved=True):
                finalized_kwargs[k] = stack.enter_context(
                    v.get(output=output, managed=managed)
                )

            with self.handle_exit(output):
                callable = cast(Callable[..., Any], self.callable)
                requires_management = inspect.isgeneratorfunction(callable)
                if requires_management:
                    # Yield functions are assumed to be context-manager style generators
                    # that we just need to wrap...
                    callable = contextlib.contextmanager(callable)

                result: Any = callable(*args, *self.args, **finalized_kwargs)
                is_context_manager = isinstance(
                    result, contextlib.AbstractContextManager
                )

                # And then enter before producing the result.
                if requires_management or (managed and is_context_manager):
                    result = stack.enter_context(result)  # pyright: ignore

            self.result = result
            yield result

    @contextlib.asynccontextmanager
    async def get_async(
        self, *args: Any, output: Output | None = None, managed: bool = True
    ) -> AsyncGenerator[C, None]:
        """Get the resolved value, in an async context.

        Note, this is the exact same process as in `get`, except with `await`,
        `enter_async_context` and `async with`. There seems to be no way to
        share the logic between the two methods, so they just need to be kept
        in sync :shrug:.
        """
        if self.result is not Empty:
            yield self.result
            return

        async with contextlib.AsyncExitStack() as stack:
            finalized_kwargs = dict(self.iter_kwargs(is_resolved=False))
            for k, v in self.iter_kwargs(is_resolved=True):
                finalized_kwargs[k] = await stack.enter_async_context(
                    v.get_async(output=output, managed=managed)
                )

            with self.handle_exit(output):
                callable = cast(Callable[..., Any], self.callable)
                requires_management = inspect.isasyncgenfunction(callable)
                if requires_management:
                    callable = contextlib.asynccontextmanager(callable)

                result: Any = callable(*args, *self.args, **finalized_kwargs)
                is_async_context_manager = isinstance(
                    result, contextlib.AbstractAsyncContextManager
                )
                is_sync_context_manager = isinstance(
                    result, contextlib.AbstractContextManager
                )

                if requires_management or (managed and is_async_context_manager):
                    result = await stack.enter_async_context(result)  # pyright: ignore
                elif managed and is_sync_context_manager:
                    # Handle synchronous context managers in async context
                    result = stack.enter_context(result)  # pyright: ignore
                elif isinstance(result, Coroutine):
                    result = await result  # pyright: ignore

            self.result = result
            yield result

    def iter_kwargs(self, *, is_resolved: bool):
        for k, v in self.kwargs.items():
            if is_resolved == isinstance(v, self.__class__):
                yield k, v

    @classmethod
    @contextlib.contextmanager
    def handle_exit(cls, output: Output | None = None):
        try:
            yield
        except Exit as e:
            if output:  # pragma: no cover
                output.exit(e)
            raise e



from __future__ import annotations

import enum
import inspect
from dataclasses import dataclass
from types import MethodType
from typing import Any, Protocol, TypeVar

from typing_extensions import assert_never

from cappa.type_view import TypeView


class DocType(Protocol):
    documentation: str


try:
    from typing_extensions import Doc as TypingExtensionsDoc

    Doc: type[DocType] = TypingExtensionsDoc
except ImportError:  # pragma: no cover

    @dataclass
    class InternalDoc:
        documentation: str

    Doc = InternalDoc


__all__ = [
    "Doc",
    "T",
    "assert_never",
    "assert_type",
    "detect_choices",
    "find_annotations",
]


T = TypeVar("T")


def find_annotations(type_view: TypeView[Any], kind: type[T]) -> list[T]:
    result: list[T] = []
    for annotation in type_view.metadata:
        if isinstance(annotation, kind):
            result.append(annotation)

        if isinstance(annotation, type) and issubclass(annotation, kind):
            result.append(annotation())

    return result


def assert_type(value: Any, typ: type[T]) -> T:
    assert isinstance(value, typ), value
    return value


def detect_choices(type_view: TypeView[Any]) -> list[str] | None:
    if type_view.is_optional:
        type_view = type_view.strip_optional()

    if type_view.is_subclass_of(enum.Enum):
        return [v.value for v in type_view.annotation]

    if type_view.is_subclass_of((list, set)) or type_view.is_variadic_tuple:
        type_view = type_view.inner_types[0]

    if type_view.is_union:
        if all(t.is_literal for t in type_view.inner_types):
            return [str(t.args[0]) for t in type_view.inner_types]

    if type_view.is_literal:
        return [str(t) for t in type_view.args]

    return None


def get_method_class(fn: MethodType) -> type:
    return inspect._findclass(fn)  # type: ignore



from __future__ import annotations

from dataclasses import dataclass

from cappa.docstring import get_attribute_docstrings


def test_attribute_docstring_single_line():
    @dataclass
    class SingleLineArgs:
        foo: int
        """A simple description."""

    result = get_attribute_docstrings(SingleLineArgs)
    assert result == {"foo": "A simple description."}


def test_attribute_docstring_indented_multiline():
    @dataclass
    class IndentedMultilineArgs:
        foo: int
        """First paragraph.

        Second paragraph.
        """

    result = get_attribute_docstrings(IndentedMultilineArgs)
    assert result == {"foo": "First paragraph.\n\nSecond paragraph."}


def test_attribute_docstring_indented_multiline_dedented():
    """Indented attribute docstrings must not be treated as Markdown code blocks."""

    @dataclass
    class IndentedDedentedArgs:
        foo: int
        """Summary line.

        Body paragraph that would have 8 spaces of indent before cleandoc,
        with a long enough body to wrap.
        """

    result = get_attribute_docstrings(IndentedDedentedArgs)

    # No leading spaces on continuation lines
    assert not any(line.startswith("    ") for line in result["foo"].splitlines())


def test_attribute_docstring_multiple_fields():
    @dataclass
    class MultipleFieldArgs:
        foo: int
        """Foo help.

        Foo body.
        """

        bar: str
        """Bar help."""

    result = get_attribute_docstrings(MultipleFieldArgs)
    assert result["foo"] == "Foo help.\n\nFoo body."
    assert result["bar"] == "Bar help."



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_negative_int_option(backend: Backend):
    @dataclass
    class Args:
        count: Annotated[int, cappa.Arg(long=True)]

    result = parse(Args, "--count", "-5", backend=backend)
    assert result == Args(count=-5)


@backends
def test_negative_float_option(backend: Backend):
    @dataclass
    class Args:
        value: Annotated[float, cappa.Arg(long=True)]

    result = parse(Args, "--value", "-3.14", backend=backend)
    assert result == Args(value=-3.14)


@backends
def test_negative_int_short_option(backend: Backend):
    @dataclass
    class Args:
        count: Annotated[int, cappa.Arg(short=True)]

    result = parse(Args, "-c", "-10", backend=backend)
    assert result == Args(count=-10)



from __future__ import annotations

from dataclasses import dataclass
from typing import Union

import pytest
from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_sequence_unioned_with_scalar(backend: Backend):
    """Inferred num_args: union arity mismatch caught in NumArgs.infer."""

    @dataclass
    class Args:
        foo: Union[list[str], str]

    with pytest.raises(ValueError) as e:
        parse(Args, "--help", backend=backend)

    exc = str(e.value).replace("typing.List", "list")
    assert exc == (
        "On field 'foo', mismatch of arity between union variants. "
        "`list[str]` produces `num_args=-1`, `<class 'str'>` produces `num_args=1`."
    )


@backends
def test_sequence_unioned_with_scalar_explicit_num_args(backend: Backend):
    """Explicit num_args bypasses NumArgs.infer union check; verify_type_compatibility catches it."""

    @dataclass
    class Args:
        foo: Annotated[Union[list[str], str], cappa.Arg(num_args=cappa.NumArgs(n=-1))]

    with pytest.raises(ValueError) as e:
        parse(Args, "--help", backend=backend)

    assert str(e.value) == (
        "On field 'foo', apparent mismatch of annotated type with `Arg` options. "
        'Unioning "sequence" types with non-sequence types is not currently supported, '
        "unless using `Arg(parse=...)` or `Arg(action=<callable>)`. "
        "See [documentation](https://cappa.readthedocs.io/en/latest/annotation.html) for more details."
    )


@backends
def test_sequence_with_scalar_action(backend: Backend):
    @dataclass
    class Args:
        foo: Annotated[list[str], cappa.Arg(action=cappa.ArgAction.set, num_args=1)]

    with pytest.raises(ValueError) as e:
        parse(Args, "--help", backend=backend)

    result = str(e.value).replace("List", "list")
    assert result == (
        "On field 'foo', apparent mismatch of annotated type with `Arg` options. "
        "'list[str]' type produces a sequence, whereas `num_args=1`/`action=ArgAction.set` do not. "
        "See [documentation](https://cappa.readthedocs.io/en/latest/annotation.html) for more details."
    )


@backends
def test_sequence_with_scalar_num_args(backend: Backend):
    @dataclass
    class Args:
        foo: Annotated[list[str], cappa.Arg(num_args=1, short=True)]

    args = parse(Args, "-f", "a", "-f", "b", backend=backend)
    assert args == Args(["a", "b"])

    @dataclass
    class ArgsBad:
        foo: Annotated[
            list[str], cappa.Arg(num_args=1, short=True, action=cappa.ArgAction.set)
        ]

    with pytest.raises(ValueError) as e:
        parse(ArgsBad, "--help", backend=backend)

    result = str(e.value).replace("List", "list")
    assert result == (
        "On field 'foo', apparent mismatch of annotated type with `Arg` options. "
        "'list[str]' type produces a sequence, whereas `num_args=1`/`action=ArgAction.set` do not. "
        "See [documentation](https://cappa.readthedocs.io/en/latest/annotation.html) for more details."
    )


@backends
def test_scalar_with_sequence_action(backend: Backend):
    @dataclass
    class Args:
        foo: Annotated[str, cappa.Arg(action=cappa.ArgAction.append)]

    with pytest.raises(ValueError) as e:
        parse(Args, "--help", backend=backend)

    assert str(e.value) == (
        "On field 'foo', apparent mismatch of annotated type with `Arg` options. "
        "'str' type produces a scalar, whereas `num_args=1`/`action=ArgAction.append` do not. "
        "See [documentation](https://cappa.readthedocs.io/en/latest/annotation.html) for more details."
    )


@backends
def test_scalar_with_sequence_num_args(backend: Backend):
    @dataclass
    class Args:
        foo: Annotated[str, cappa.Arg(num_args=5)]

    with pytest.raises(ValueError) as e:
        parse(Args, "--help", backend=backend)

    assert str(e.value) == (
        "On field 'foo', apparent mismatch of annotated type with `Arg` options. "
        "'str' type produces a scalar, whereas `num_args=5`/`action=ArgAction.set` do not. "
        "See [documentation](https://cappa.readthedocs.io/en/latest/annotation.html) for more details."
    )


@backends
def test_num_args_optional_value_requires_optional_type(backend: Backend):
    """NumArgs(required=False, default=None) must pair with an Optional type."""

    @dataclass
    class Args:
        foo: Annotated[
            int, cappa.Arg(long=True, num_args=cappa.NumArgs(required=False))
        ] = 5

    with pytest.raises(ValueError) as e:
        parse(Args, "--help", backend=backend)

    assert str(e.value) == (
        "On field 'foo', `NumArgs(required=False, default=None)` requires the "
        "annotated type to include `None` (e.g. `int | None`), "
        "since the absent-value default of `None` is not assignable to the field type."
    )


@backends
def test_num_args_optional_value_non_none_default_does_not_require_optional_type(
    backend: Backend,
):
    """NumArgs(required=False, default=0) is valid with a non-optional type."""

    @dataclass
    class Args:
        foo: Annotated[
            int, cappa.Arg(long=True, num_args=cappa.NumArgs(required=False, default=0))
        ] = 5

    # Should not raise — default is a valid int, type need not include None.
    args = parse(Args, "--foo", backend=backend)
    assert args == Args(foo=0)



from __future__ import annotations

from dataclasses import dataclass

import pytest
from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_arg_name_disallowed(backend: Backend):
    @dataclass
    class ArgTest:
        bad: Annotated[bool, cappa.Arg(field_name="oops")] = False

    with pytest.raises(
        ValueError, match="Arg 'name' cannot be set when using automatic inference."
    ):
        parse(ArgTest, backend=backend)



from __future__ import annotations

import dataclasses
import re
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from cappa.arg import Arg


@dataclasses.dataclass
class ShellHandler:
    name: str
    template: str

    def backend_template(self, prog: str, arg: Arg[Any]) -> str:
        safe_name = re.sub(r"\W*", "", prog.replace("-", "_"), flags=re.ASCII)

        assert isinstance(arg.long, list)
        assert len(arg.long) > 0
        return self.template % {
            "prog_name": prog,
            "safe_prog_name": safe_name,
            "completion_arg": arg.long[0],
        }


@dataclasses.dataclass
class Completion:
    value: str | None = None
    help: str | None = None
    arg: Arg[Any] | None = None

    @property
    def description(self) -> str:
        if not self.arg or self.arg.help:
            return self.help or ""

        return f"<{self.arg.value_name}> {self.help}".strip()


@dataclasses.dataclass
class FileCompletion:
    text: str



from cappa.default import Confirm, Default, Env, Prompt, ValueFrom
from cappa.type_view import Empty


def test_default_combination():
    assert Default().fallback_to(Default()) == Default()
    assert (Default() | Default()) == Default()

    assert Env("FOO") | Env("BAR") == Default(Env("FOO"), Env("BAR"), default=Empty)
    assert Default(default=4) | Env("FOO") == Default(Env("FOO"), default=4)
    assert Env("FOO") | ValueFrom(int) == Default(Env("FOO"), ValueFrom(int))

    prompt = Prompt("FOO")
    confirm = Confirm("BAR")
    assert Default() | prompt | confirm == Default(prompt, confirm)


def test_default_eval():
    default = Default()
    assert default() == (True, None)

    default = Default(Env("FOO"))
    assert default() == (True, None)

    default = Default(Env("FOO", default="foo"))
    assert default() == (False, "foo")

    default = Default(Env("FOO"), Env("BAR"), Env("BAZ", default="okay"))
    assert default() == (False, "okay")


def test_value_from():
    def fn():
        return 5

    default = ValueFrom(fn)
    assert default() == 5

    def fn2(value: int):
        return value + 5

    default = ValueFrom(fn2, value=4)
    assert default() == 9

    def fn3(*, value: int):
        return value + 5

    default = ValueFrom(fn3, value=4)
    assert default() == 9



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

import cappa
from tests.utils import Backend, backends, invoke


def foo(command: Command):
    print("called", command)


def bar():
    print("two")


def command():
    return 1


@cappa.command(invoke=command)
@dataclass
class Command: ...


@backends
def test_invoke_global_dep(capsys: Any, backend: Backend):
    result = invoke(Command, deps=[foo, bar], backend=backend)
    assert result == 1

    out = capsys.readouterr().out
    assert "called Command()" in out
    assert "two" in out



from __future__ import annotations

from dataclasses import dataclass
from typing import Union

import pytest
from typing_extensions import Annotated

import cappa
from tests.utils import Backend, backends, parse


@backends
def test_(backend: Backend):
    @dataclass
    class Args:
        foo: Annotated[str, cappa.Arg(long=True)]
        raw: Annotated[Union[list[str], None], cappa.Arg(num_args=-1)] = None

    with pytest.raises(cappa.Exit) as e:
        parse(Args, "--foo", "--", "value", backend=backend)

    message = str(e.value.message)
    if backend:
        assert "--foo: expected one argument" in message
    else:
        assert "Option 'foo' requires an argument" in message
